import warnings

import rasterio


def rasterio_decorator(func):
    def wrapped_f(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with rasterio.drivers():
                return func(*args, **kwargs)

    return wrapped_f




geojson_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "id": "http://json-schema.org/geojson/geojson.json#",
    "title": "Geo JSON object",
    "description": "Schema for a Geo JSON object",
    "type": "object",
    "required": ["type"],
    "properties": {
        "crs": {"$ref": "http://json-schema.org/geojson/crs.json#"},
        "bbox": {"$ref": "http://json-schema.org/geojson/bbox.json#"}
    },
    "oneOf": [
        {"$ref": "#/definitions/geometry"},
        {"$ref": "#/definitions/geometryCollection"},
        {"$ref": "#/definitions/feature"},
        {"$ref": "#/definitions/featureCollection"}
    ],
    "definitions": {
        "position": {
            "description": "A single position",
            "type": "array",
            "minItems": 2,
            "items": [ {"type": "number"}, {"type": "number"} ],
            "additionalItems": False
        },
        "positionArray": {
            "description": "An array of positions",
            "type": "array",
            "items": {"$ref": "#/definitions/position"}
        },
        "lineString": {
            "description": "An array of two or more positions",
            "allOf": [
                {"$ref": "#/definitions/positionArray"},
                {"minItems": 2}
            ]
        },
        "linearRing": {
            "description": "An array of four positions where the first equals the last",
            "allOf": [
                {"$ref": "#/definitions/positionArray"},
                {"minItems": 4}
            ]
        },
        "polygon": {
            "description": "An array of linear rings",
            "type": "array",
            "items": {"$ref": "#/definitions/linearRing"}
        },
        "geometry": {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "geometry",
            "description": "One geometry as defined by GeoJSON",
            "type": "object",
            "required": ["type", "coordinates"],
            "oneOf": [
                {
                    "title": "Point",
                    "properties": {
                        "type": {"enum": ["Point"]},
                        "coordinates": {"$ref": "#/definitions/position"}
                    }
                },
                {
                    "title": "MultiPoint",
                    "properties": {
                        "type": {"enum": ["MultiPoint"]},
                        "coordinates": {"$ref": "#/definitions/positionArray"}
                    }
                },
                {
                    "title": "LineString",
                    "properties": {
                        "type": {"enum": ["LineString"]},
                        "coordinates": {"$ref": "#/definitions/lineString"}
                    }
                },
                {
                    "title": "MultiLineString",
                    "properties": {
                        "type": {"enum": [ "MultiLineString" ]},
                        "coordinates": {
                            "type": "array",
                            "items": {"$ref": "#/definitions/lineString"}
                        }
                    }
                },
                {
                    "title": "Polygon",
                    "properties": {
                        "type": {"enum": [ "Polygon" ]},
                        "coordinates": {"$ref": "#/definitions/polygon"}
                    }
                },
                {
                    "title": "MultiPolygon",
                    "properties": {
                        "type": {"enum": ["MultiPolygon"]},
                        "coordinates": {
                            "type": "array",
                            "items": {"$ref": "#/definitions/polygon"}
                        }
                    }
                }
            ]
        },
        "geometryCollection": {
            "title": "GeometryCollection",
            "description": "A collection of geometry objects",
            "required": [ "geometries" ],
            "properties": {
                "type": {"enum": [ "GeometryCollection" ]},
                "geometries": {
                    "type": "array",
                    "items": {"$ref": "#/definitions/geometry"}
                }
            }
        },
        "feature": {
            "title": "Feature",
            "description": "A Geo JSON feature object",
            "required": [ "geometry", "properties" ],
            "properties": {
                "type": {"enum": [ "Feature" ]},
                "geometry": {
                    "oneOf": [
                        {"type": "null"},
                        {"$ref": "#/definitions/geometry"}
                    ]
                },
                "properties": {"type": [ "object", "null" ]},
                "id": {"FIXME": "may be there, type not known (string? number?)"}
            }
        },
        "featureCollection": {
            "title": "FeatureCollection",
            "description": "A Geo JSON feature collection",
            "required": [ "features" ],
            "properties": {
                "type": {"enum": [ "FeatureCollection" ]},
                "features": {
                    "type": "array",
                    "items": {"$ref": "#/definitions/feature"}
                }
            }
        }
    }
}



from __future__ import print_function, division, absolute_import

from os.path import join

import rasterio
import numpy

from . import settings
from .decorators import rasterio_decorator
from .image import BaseProcess


class NDVI(BaseProcess):

    def __init__(self, path, bands=None, **kwargs):
        bands = [4, 5]
        self._read_cmap()
        super(NDVI, self).__init__(path, bands, **kwargs)

    def _read_cmap(self):
        """
        reads the colormap from a text file given in settings.py.
        See colormap_cubehelix.txt. File must contain 256 RGB values
        """

        try:
            i = 0
            colormap = {0: (0, 0, 0)}
            with open(settings.COLORMAP) as cmap:
                lines = cmap.readlines()
                for line in lines:
                    if i == 0 and 'mode = ' in line:
                        i = 1
                        maxval = float(line.replace('mode = ', ''))
                    elif i > 0:
                        str = line.split()
                        if str == []:  # when there are empty lines at the end of the file
                            break
                        colormap.update(
                            {
                                i: (int(round(float(str[0]) * 255 / maxval)),
                                    int(round(float(str[1]) * 255 / maxval)),
                                    int(round(float(str[2]) * 255 / maxval)))
                            }
                        )
                        i += 1
        except IOError:
            pass

        self.cmap = {k: v[:4] for k, v in colormap.items()}

    @rasterio_decorator
    def run(self):
        """
        Executes NDVI processing
        """
        self.output("* NDVI processing started.", normal=True)

        bands = self._read_bands()
        image_data = self._get_image_data()

        new_bands = []
        for i in range(0, 2):
            new_bands.append(numpy.empty(image_data['shape'], dtype=numpy.float32))

        self._warp(image_data, bands, new_bands)

        # Bands are no longer needed
        del bands

        calc_band = numpy.true_divide((new_bands[1] - new_bands[0]), (new_bands[1] + new_bands[0]))

        output_band = numpy.rint((calc_band + 1) * 255 / 2).astype(numpy.uint8)

        output_file = join(self.dst_path, self._filename(suffix='NDVI'))

        return self.write_band(output_band, output_file, image_data)

    def write_band(self, output_band, output_file, image_data):

        # from http://publiclab.org/notes/cfastie/08-26-2014/new-ndvi-colormap
        with rasterio.open(output_file, 'w', driver='GTiff',
                           width=image_data['shape'][1],
                           height=image_data['shape'][0],
                           count=1,
                           dtype=numpy.uint8,
                           nodata=0,
                           transform=image_data['dst_transform'],
                           crs=self.dst_crs) as output:

            output.write_band(1, output_band)

            self.output("Writing to file", normal=True, color='green', indent=1)
        return output_file


class NDVIWithManualColorMap(NDVI):

    def manual_colormap(self, n, i):
        return self.cmap[n][i]

    def write_band(self, output_band, output_file, image_data):
        # colormaps will overwrite our transparency masks so we will manually
        # create three RGB bands

        self.output("Applying ColorMap", normal=True, arrow=True)
        self.cmap[0] = (0, 0, 0, 255)

        v_manual_colormap = numpy.vectorize(self.manual_colormap, otypes=[numpy.uint8])
        rgb_bands = []
        for i in range(3):
            rgb_bands.append(v_manual_colormap(output_band, i))

        with rasterio.drivers(GDAL_TIFF_INTERNAL_MASK=True):
            with rasterio.open(output_file, 'w', driver='GTiff',
                               width=image_data['shape'][1],
                               height=image_data['shape'][0],
                               count=3,
                               dtype=numpy.uint8,
                               nodata=0,
                               photometric='RGB',
                               transform=image_data['dst_transform'],
                               crs=self.dst_crs) as output:

                for i in range(3):
                    output.write_band(i + 1, rgb_bands[i])

            self.output("Writing to file", normal=True, color='green', indent=1)
        return output_file



# Boto Uploader
# Landsat Util
# License: CC0 1.0 Universal

# The S3 uploader is a fork of pys3upload (https://github.com/leetreveil/pys3upload)

from __future__ import print_function, division, absolute_import

import os
import sys
import time
import threading
import contextlib

try:
    import queue
except:
    import Queue as queue

from multiprocessing import pool

try:
    from io import BytesIO as StringIO
except ImportError:
    try:
        from cStringIO import StringIO
    except:
        from StringIO import StringIO

from boto.s3.connection import S3Connection

from .mixins import VerbosityMixin

STREAM = sys.stderr


class Uploader(VerbosityMixin):

    """
    The Uploader class.

    To initiate the following parameters must be passed:

    :param key:
        AWS access key id (optional)
    :type key:
        String
    :param secret:
        AWS access secret key (optional)
    :type secret:
        String
    :param host:
        AWS host, e.g. s3.amazonaws.com (optional)
    :type host:
        String
    """

    progress_template = \
        'File Size:%(size)4d MB | Uploaded:%(uploaded)4d MB' + ' ' * 8

    def __init__(self, key=None, secret=None, host=None):
        self.key = key
        self.secret = secret
        self.source_size = 0
        self.conn = S3Connection(key, secret, host=host)

    def run(self, bucket_name, filename, path):
        """
        Initiate the upload.

        :param bucket_name:
            Name of the S3 bucket
        :type bucket_name:
            String
        :param filename:
            The filname
        :type filename:
            String
        :param path:
            The path to the file that needs to be uploaded
        :type path:
            String

        :returns:
            void
        """

        f = open(path, 'rb')
        self.source_size = os.stat(path).st_size
        total_dict = {}

        def cb(part_no, uploaded, total):

            total_dict[part_no] = uploaded

            params = {
                'uploaded': round(sum(total_dict.values()) / 1048576, 0),
                'size': round(self.source_size / 1048576, 0),
            }

            p = (self.progress_template + '\r') % params

            STREAM.write(p)
            STREAM.flush()

        self.output('Uploading to S3', normal=True, arrow=True)
        upload(bucket_name, self.key, self.secret,
               data_collector(iter(f)), filename, cb,
               threads=10, replace=True, secure=True, connection=self.conn)

        print('\n')
        self.output('Upload Completed', normal=True, arrow=True)


def data_collector(iterable, def_buf_size=5242880):
    """ Buffers n bytes of data.

    :param iterable:
        Could be a list, generator or string
    :type iterable:
        List, generator, String

    :returns:
        A generator object
    """
    buf = b''
    for data in iterable:
        buf += data
        if len(buf) >= def_buf_size:
            output = buf[:def_buf_size]
            buf = buf[def_buf_size:]
            yield output
    if len(buf) > 0:
        yield buf


def upload_part(upload_func, progress_cb, part_no, part_data):
    num_retries = 5

    def _upload_part(retries_left=num_retries):
        try:
            with contextlib.closing(StringIO(part_data)) as f:
                f.seek(0)
                cb = lambda c, t: progress_cb(part_no, c, t) if progress_cb else None
                upload_func(f, part_no, cb=cb, num_cb=100)
        except Exception as exc:
            retries_left -= 1
            if retries_left > 0:
                return _upload_part(retries_left=retries_left)
            else:
                return threading.ThreadError(repr(threading.current_thread()) + ' ' + repr(exc))
    return _upload_part()


def upload(bucket, aws_access_key, aws_secret_key,
           iterable, key, progress_cb=None,
           threads=5, replace=False, secure=True,
           connection=None):
    """ Upload data to s3 using the s3 multipart upload API.

    :param bucket:
        Name of the S3 bucket
    :type bucket:
        String
    :param aws_access_key:
        AWS access key id (optional)
    :type aws_access_key:
        String
    :param aws_secret_key:
        AWS access secret key (optional)
    :type aws_secret_key:
        String
    :param iterable:
        The data to upload. Each 'part' in the list. will be uploaded in parallel. Each part must be at
        least 5242880 bytes (5mb).
    :type iterable:
        An iterable object
    :param key:
        The name of the key (filename) to create in the s3 bucket
    :type key:
        String
    :param progress_cb:
        Progress callback, will be called with (part_no, uploaded, total) each time a progress update
        is available. (optional)
    :type progress_cb:
        function
    :param threads:
        the number of threads to use while uploading. (Default is 5)
    :type threads:
        int
    :param replace:
        will replace the key (filename) on S3 if set to true. (Default is false)
    :type replace:
        boolean
    :param secure:
        Use ssl when talking to s3. (Default is true)
    :type secure:
        boolean
    :param connection:
        Used for testing (optional)
    :type connection:
        S3 connection class

    :returns:
        void
    """

    if not connection:
        from boto.s3.connection import S3Connection as connection
        c = connection(aws_access_key, aws_secret_key, is_secure=secure)
    else:
        c = connection

    b = c.get_bucket(bucket)

    if not replace and b.lookup(key):
        raise Exception('s3 key ' + key + ' already exists')

    multipart_obj = b.initiate_multipart_upload(key)
    err_queue = queue.Queue()
    lock = threading.Lock()
    upload.counter = 0

    try:
        tpool = pool.ThreadPool(processes=threads)

        def check_errors():
            try:
                exc = err_queue.get(block=False)
            except queue.Empty:
                pass
            else:
                raise exc

        def waiter():
            while upload.counter >= threads:
                check_errors()
                time.sleep(0.1)

        def cb(err):
            if err:
                err_queue.put(err)
            with lock:
                upload.counter -= 1

        args = [multipart_obj.upload_part_from_file, progress_cb]

        for part_no, part in enumerate(iterable):
            part_no += 1
            tpool.apply_async(upload_part, args + [part_no, part], callback=cb)
            with lock:
                upload.counter += 1
            waiter()

        tpool.close()
        tpool.join()
        # Check for thread errors before completing the upload,
        # sometimes an error can be left unchecked until we
        # get to this point.
        check_errors()
        multipart_obj.complete_upload()
    except:
        multipart_obj.cancel_upload()
        tpool.terminate()
        raise



# Landsat Util
# License: CC0 1.0 Universal

##
# Main Setting File
##

from os import getenv
from os.path import join, expanduser, abspath, dirname

# Google Storage Landsat Config

DEBUG = getenv('DEBUG', False)

SATELLITE = 'L8'
L8_METADATA_URL = 'http://landsat.usgs.gov/metadata_service/bulk_metadata_files/LANDSAT_8.csv'
GOOGLE_STORAGE = 'http://storage.googleapis.com/earthengine-public/landsat/'
S3_LANDSAT = 'http://landsat-pds.s3.amazonaws.com/'
API_URL = 'https://api.developmentseed.org/satellites/landsat'

# User's Home Directory
HOME_DIR = expanduser('~')

# Utility's base directory
BASE_DIR = abspath(dirname(__file__))

LANDSAT_DIR = join(HOME_DIR, 'landsat')
DOWNLOAD_DIR = join(LANDSAT_DIR, 'downloads')
PROCESSED_IMAGE = join(LANDSAT_DIR, 'processed')

# Colormap File
COLORMAP = join(abspath(dirname(__file__)), 'maps', 'colormap_ndvi_cfastie.txt')



# Landsat Util
# License: CC0 1.0 Universal

"""Tests for mixins"""

from __future__ import absolute_import

import sys
import unittest

try:
    from io import StringIO
except:
    from cStringIO import StringIO

from contextlib import contextmanager

from landsat.mixins import VerbosityMixin


# Capture function is taken from
# http://schinckel.net/2013/04/15/capture-and-test-sys.stdout-sys.stderr-in-unittest.testcase/
@contextmanager
def capture(command, *args, **kwargs):
    out, sys.stdout = sys.stdout, StringIO()
    command(*args, **kwargs)
    sys.stdout.seek(0)
    yield sys.stdout.read()
    sys.stdout = out


class TestMixins(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.v = VerbosityMixin()

    def test_output(self):
        # just a value
        with capture(self.v.output, u'this is a test') as output:
            self.assertEquals("", output)

        # value as normal
        with capture(self.v.output, u'this is a test', normal=True) as output:
            self.assertEquals("this is a test\n", output)

        # value as normal with color
        with capture(self.v.output, u'this is a test', normal=True, color='blue') as output:
            self.assertEquals("\x1b[34mthis is a test\x1b[0m\n", output)

        # value as error
        with capture(self.v.output, u'this is a test', normal=True, error=True) as output:
            self.assertEquals("\x1b[31mthis is a test\x1b[0m\n", output)

        # value with arrow
        with capture(self.v.output, u'this is a test', normal=True, arrow=True) as output:
            self.assertEquals("\x1b[34m===> \x1b[0mthis is a test\n", output)

        # value with indent
        with capture(self.v.output, u'this is a test', normal=True, indent=1) as output:
            self.assertEquals("     this is a test\n", output)

    def test_exit(self):
        with self.assertRaises(SystemExit):
            with capture(self.v.exit, u'exit test') as output:
                self.assertEquals('exit test', output)

    def test_print(self):
        # message in blue with arrow
        with capture(self.v._print, msg=u'this is a test', color='blue', arrow=True) as output:
            self.assertEquals("\x1b[34m===> \x1b[0m\x1b[34mthis is a test\x1b[0m\n", output)

        # just a message
        with capture(self.v._print, msg=u'this is a test') as output:
            self.assertEquals("this is a test\n", output)

        # message with color and indent
        with capture(self.v._print, msg=u'this is a test', color='blue', indent=1) as output:
            self.assertEquals("     \x1b[34mthis is a test\x1b[0m\n", output)


if __name__ == '__main__':
    unittest.main()



# Landsat Util
# License: CC0 1.0 Universal

"""Tests for image processing"""

from os.path import join, abspath, dirname, exists
import errno
import shutil
import unittest
from tempfile import mkdtemp

import rasterio
from rasterio.warp import transform_bounds

from landsat.image import Simple, PanSharpen
from landsat.ndvi import NDVI, NDVIWithManualColorMap


def get_bounds(path):
    """ Retrun bounds in WGS84 system """

    with rasterio.drivers():
        src = rasterio.open(path)

        return transform_bounds(
                    src.crs,
                    {'init': 'EPSG:4326'},
                    *src.bounds)


class TestProcess(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.base_dir = abspath(dirname(__file__))
        cls.temp_folder = mkdtemp()
        cls.landsat_image = join(cls.base_dir, 'samples/test.tar.bz2')

    @classmethod
    def tearDownClass(cls):

        try:
            shutil.rmtree(cls.temp_folder)
            shutil.rmtree(join(cls.base_dir, 'samples', 'test'))
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise

    def test_simple_no_bands(self):

        p = Simple(path=self.landsat_image, dst_path=self.temp_folder)
        self.assertTrue(exists(p.run()))

    def test_simple_with_bands(self):

        p = Simple(path=self.landsat_image, bands=[1, 2, 3], dst_path=self.temp_folder)
        self.assertTrue(exists(p.run()))

    def test_simple_with_clip(self):

        bounds = [-87.48138427734375, 30.700515832683923, -87.43331909179688, 30.739475058679485]
        p = Simple(path=self.landsat_image, bands=[1, 2, 3], dst_path=self.temp_folder,
                   bounds=bounds)
        path = p.run()
        self.assertTrue(exists(path))
        for val, exp in zip(get_bounds(path), bounds):
            self.assertAlmostEqual(val, exp, 2)

    def test_simple_with_intersecting_bounds_clip(self):

        bounds = [-87.520515832683923, 30.700515832683923, -87.43331909179688, 30.739475058679485]
        expected_bounds = [-87.49691403528307, 30.700515832683923, -87.43331909179688, 30.739475058679485]
        p = Simple(path=self.landsat_image, bands=[1, 2, 3], dst_path=self.temp_folder,
                   bounds=bounds)
        path = p.run()
        self.assertTrue(exists(path))
        for val, exp in zip(get_bounds(path), expected_bounds):
            self.assertAlmostEqual(val, exp, 2)

    def test_simple_with_out_of_bounds_clip(self):

        bounds = [-87.66197204589844, 30.732392734006083, -87.57545471191406, 30.806731169315675]
        expected_bounds = [-87.49691403528307, 30.646646570857722, -87.29976764207227, 30.810617911193567]
        p = Simple(path=self.landsat_image, bands=[1, 2, 3], dst_path=self.temp_folder,
                   bounds=bounds)
        path = p.run()
        self.assertTrue(exists(path))
        for val, exp in zip(get_bounds(path), expected_bounds):
            self.assertAlmostEqual(val, exp, 2)

    def test_simple_with_zip_file(self):

        p = Simple(path=self.landsat_image, dst_path=self.temp_folder)

        # test from an unzip file
        self.path = join(self.base_dir, 'samples', 'test')
        self.assertTrue(exists(p.run()))

    def test_pansharpen(self):
        p = PanSharpen(path=self.landsat_image, bands=[4, 3, 2], dst_path=self.temp_folder)
        self.assertTrue(exists(p.run()))

    def test_pansharpen_with_clip(self):
        """ test with pansharpen and clipping """

        bounds = [-87.48138427734375, 30.700515832683923, -87.43331909179688, 30.739475058679485]
        p = PanSharpen(path=self.landsat_image, bands=[4, 3, 2], dst_path=self.temp_folder,
                       bounds=bounds)
        path = p.run()
        self.assertTrue(exists(path))
        for val, exp in zip(get_bounds(path), bounds):
            self.assertAlmostEqual(val, exp, 2)

    def test_ndvi(self):

        p = NDVI(path=self.landsat_image, dst_path=self.temp_folder)
        self.assertTrue(exists(p.run()))

    def test_ndvi_with_clip(self):

        bounds = [-87.48138427734375, 30.700515832683923, -87.43331909179688, 30.739475058679485]
        p = NDVI(path=self.landsat_image, dst_path=self.temp_folder,
                 bounds=bounds)
        path = p.run()
        self.assertTrue(exists(path))
        for val, exp in zip(get_bounds(path), bounds):
            self.assertAlmostEqual(val, exp, 2)

    def test_ndvi_with_manual_colormap(self):

        p = NDVIWithManualColorMap(path=self.landsat_image, dst_path=self.temp_folder)
        self.assertTrue(exists(p.run()))



# Landsat Util
# License: CC0 1.0 Universal

from __future__ import print_function, division, absolute_import

import os
import sys
import time
import re

try:
    from io import StringIO
except ImportError:
    from cStringIO import StringIO

from datetime import datetime

import geocoder

from .mixins import VerbosityMixin


class Capturing(list):
    """
    Captures a subprocess stdout.

    :Usage:
        >>> with Capturing():
        ...     subprocess(args)
    """
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        sys.stdout = self._stdout


class timer(object):
    """
    A timer class.

    :Usage:
        >>> with timer():
        ...     your code
    """
    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        self.end = time.time()
        print('Time spent : {0:.2f} seconds'.format((self.end - self.start)))


def exit(message, code=0):
    """ output a message to stdout and terminates the process.

    :param message:
        Message to be outputed.
    :type message:
        String
    :param code:
        The termination code. Default is 0
    :type code:
        int

    :returns:
        void
    """

    v = VerbosityMixin()
    if code == 0:
        v.output(message, normal=True, arrow=True)
        v.output('Done!', normal=True, arrow=True)
    else:
        v.output(message, normal=True, error=True)
    sys.exit(code)


def create_paired_list(value):
    """ Create a list of paired items from a string.

    :param value:
        the format must be 003,003,004,004 (commas with no space)
    :type value:
        String

    :returns:
        List

    :example:
        >>> create_paired_list('003,003,004,004')
        [['003','003'], ['004', '004']]

    """

    if isinstance(value, list):
        value = ",".join(value)

    array = re.split('\D+', value)

    # Make sure the elements in the list are even and pairable
    if len(array) % 2 == 0:
        new_array = [list(array[i:i + 2]) for i in range(0, len(array), 2)]
        return new_array
    else:
        raise ValueError('The string should include pairs and be formated. '
                         'The format must be 003,003,004,004 (commas with '
                         'no space)')


def check_create_folder(folder_path):
    """ Check whether a folder exists, if not the folder is created.

    :param folder_path:
        Path to the folder
    :type folder_path:
        String

    :returns:
        (String) the path to the folder
    """
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)

    return folder_path


def get_file(path):
    """ Separate the name of the file or folder from the path and return it.

    :param path:
        Path to the folder
    :type path:
        String

    :returns:
        (String) the filename

    :example:
        >>> get_file('/path/to/file.jpg')
        'file.jpg'
    """
    return os.path.basename(path)


def get_filename(path):
    """ Return the filename without extension.

    :param path:
        Path to the folder
    :type path:
        String

    :returns:
        (String) the filename without extension

    :example:
        >>> get_filename('/path/to/file.jpg')
        'file'
    """
    return os.path.splitext(get_file(path))[0]


def three_digit(number):
    """ Add 0s to inputs that their length is less than 3.

    :param number:
        The number to convert
    :type number:
        int

    :returns:
        String

    :example:
        >>> three_digit(1)
        '001'
    """
    number = str(number)
    if len(number) == 1:
        return u'00%s' % number
    elif len(number) == 2:
        return u'0%s' % number
    else:
        return number


def georgian_day(date):
    """ Returns the number of days passed since the start of the year.

    :param date:
        The string date with this format %m/%d/%Y
    :type date:
        String

    :returns:
        int

    :example:
        >>> georgian_day('05/1/2015')
        121
    """
    try:
        fmt = '%m/%d/%Y'
        return datetime.strptime(date, fmt).timetuple().tm_yday
    except (ValueError, TypeError):
        return 0


def year(date):
    """ Returns the year.

    :param date:
        The string date with this format %m/%d/%Y
    :type date:
        String

    :returns:
        int

    :example:
        >>> year('05/1/2015')
        2015
    """
    try:
        fmt = '%m/%d/%Y'
        return datetime.strptime(date, fmt).timetuple().tm_year
    except ValueError:
        return 0


def reformat_date(date, new_fmt='%Y-%m-%d'):
    """ Returns reformated date.

    :param date:
        The string date with this format %m/%d/%Y
    :type date:
        String
    :param new_fmt:
        date format string. Default is '%Y-%m-%d'
    :type date:
        String

    :returns:
        int

    :example:
        >>> reformat_date('05/1/2015', '%d/%m/%Y')
        '1/05/2015'
    """
    try:
        if isinstance(date, datetime):
            return date.strftime(new_fmt)
        else:
            fmt = '%m/%d/%Y'
            return datetime.strptime(date, fmt).strftime(new_fmt)
    except ValueError:
        return date


def convert_to_integer_list(value):
    """ Converts a comma separate string to a list

    :param value:
        the format must be 003,003,004,004 (commas with no space)
    :type value:
        String

    :returns:
        List

    :example:
        >>> convert_to_integer_list('003,003,004,004')
        ['003', '003', '004', '004']

    """
    if isinstance(value, list) or value is None:
        return value
    else:
        s = re.findall('(10|11|QA|[0-9])', value)
        for k, v in enumerate(s):
            try:
                s[k] = int(v)
            except ValueError:
                pass
        return s


# Geocoding confidence scores, from https://github.com/DenisCarriere/geocoder/blob/master/docs/features/Confidence%20Score.md
geocode_confidences = {
    10: 0.25,
    9: 0.5,
    8: 1.,
    7: 5.,
    6: 7.5,
    5: 10.,
    4: 15.,
    3: 20.,
    2: 25.,
    1: 99999.,
    # 0: unable to locate at all
}


def geocode(address, required_precision_km=1.):
    """ Identifies the coordinates of an address

    :param address:
        the address to be geocoded
    :type value:
        String
    :param required_precision_km:
        the maximum permissible geographic uncertainty for the geocoding
    :type required_precision_km:
        float

    :returns:
        dict

    :example:
        >>> geocode('1600 Pennsylvania Ave NW, Washington, DC 20500')
        {'lat': 38.89767579999999, 'lon': -77.0364827}

    """
    geocoded = geocoder.google(address)
    precision_km = geocode_confidences[geocoded.confidence]

    if precision_km <= required_precision_km:
        (lon, lat) = geocoded.geometry['coordinates']
        return {'lat': lat, 'lon': lon}
    else:
        raise ValueError("Address could not be precisely located")


def convert_to_float_list(value):
    """ Converts a comma separate string to a list

    :param value:
        the format must be 1.2,-3.5 (commas with no space)
    :type value:
        String

    :returns:
        List

    :example:
        >>> convert_to_integer_list('003,003,004,004')
        [1.2, -3.5]

    """
    if isinstance(value, list) or value is None:
        return value
    else:
        s = re.findall('([-+]?\d*\.\d+|\d+|[-+]?\d+)', value)
        for k, v in enumerate(s):
            try:
                s[k] = float(v)
            except ValueError:
                pass
        return s


def adjust_bounding_box(bounds1, bounds2):
    """ If the bounds 2 corners are outside of bounds1, they will be adjusted to bounds1 corners

    @params
    bounds1 - The source bounding box
    bounds2 - The target bounding box that has to be within bounds1

    @return
    A bounding box tuple in (y1, x1, y2, x2) format
    """

    # out of bound check
    # If it is completely outside of target bounds, return target bounds
    if ((bounds2[0] > bounds1[0] and bounds2[2] > bounds1[0]) or
            (bounds2[2] < bounds1[2] and bounds2[2] < bounds1[0])):
        return bounds1

    if ((bounds2[1] < bounds1[1] and bounds2[3] < bounds1[1]) or
            (bounds2[3] > bounds1[3] and bounds2[1] > bounds1[3])):
        return bounds1

    new_bounds = list(bounds2)

    # Adjust Y axis (Longitude)
    if (bounds2[0] > bounds1[0] or bounds2[0] < bounds1[3]):
        new_bounds[0] = bounds1[0]
    if (bounds2[2] < bounds1[2] or bounds2[2] > bounds1[0]):
        new_bounds[2] = bounds1[2]

    # Adjust X axis (Latitude)
    if (bounds2[1] < bounds1[1] or bounds2[1] > bounds1[3]):
        new_bounds[1] = bounds1[1]
    if (bounds2[3] > bounds1[3] or bounds2[3] < bounds1[1]):
        new_bounds[3] = bounds1[3]

    return tuple(new_bounds)


def remove_slash(value):

    assert(isinstance(value, str))
    return re.sub('(^\/|\/$)', '', value)


def url_builder(segments):

    # Only accept list or tuple
    assert((isinstance(segments, list) or isinstance(segments, tuple)))
    return "/".join([remove_slash(s) for s in segments])




# Landsat Util
# License: CC0 1.0 Universal

"""Tests for utils"""

from os.path import join
import errno
import shutil
import unittest
from tempfile import mkdtemp, mkstemp

from landsat import utils


class TestUtils(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.temp_folder_base = mkdtemp()
        cls.temp_folder_test = join(cls.temp_folder_base, 'test')
        cls.temp_file = mkstemp(dir=cls.temp_folder_base)

    @classmethod
    def tearDownClass(cls):
        try:
            shutil.rmtree(cls.temp_folder_base)
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise

    def test_create_paired_list(self):
        # Test correct input (string)
        output = utils.create_paired_list('003,003,004,004')
        self.assertEqual([['003', '003'], ['004', '004']], output)

        # Test correct input (list)
        output = utils.create_paired_list(['003', '003', '004', '004'])
        self.assertEqual([['003', '003'], ['004', '004']], output)

        # Test incorrect input
        self.assertRaises(ValueError, utils.create_paired_list, '003,003,004')
        self.assertRaises(ValueError, utils.create_paired_list, '')

    def test_check_create_folder(self):
        new_path = utils.check_create_folder(self.temp_folder_test)

        self.assertEqual(new_path, self.temp_folder_test)

    def test_get_file(self):
        f = utils.get_file(self.temp_folder_test)

        self.assertEqual('test', f)

    def test_get_filename(self):
        # Easy filename
        f = utils.get_filename('%s/filename.html' % self.temp_folder_base)
        self.assertEqual('filename', f)

        # Dificult filename
        f = utils.get_filename('%s/filename.test.html' % self.temp_folder_base)
        self.assertEqual('filename.test', f)

    def test_three_digit(self):
        self.assertEqual('009', utils.three_digit(9))
        self.assertEqual('010', utils.three_digit(10))
        self.assertEqual('100', utils.three_digit(100))
        self.assertEqual('string', utils.three_digit('string'))

    def test_georgian_day(self):
        self.assertEqual(28, utils.georgian_day('01/28/2014'))
        self.assertEqual(79, utils.georgian_day('03/20/2014'))
        self.assertEqual(0, utils.georgian_day('random text'))
        self.assertEqual(0, utils.georgian_day(9876))

    def test_year(self):
        self.assertEqual(2014, utils.year('01/28/2014'))
        self.assertEqual(0, utils.year('2014'))

    def test_reformat_date(self):

        self.assertEqual('2014-02-03', utils.reformat_date('02/03/2014'))
        self.assertEqual('2014-03-02', utils.reformat_date('02/03/2014', '%Y-%d-%m'))
        self.assertEqual('2014', utils.reformat_date('2014', '%d/%m/%Y'))
        self.assertEqual('2014', utils.reformat_date('2014', 'juberish'))
        self.assertRaises(TypeError, utils.reformat_date('date'))

    def test_convert_to_integer_list(self):
        # correct input
        r = utils.convert_to_integer_list('1,2,3')
        self.assertEqual([1, 2, 3], r)

        # try other cobinations
        r = utils.convert_to_integer_list('1, 2, 3')
        self.assertEqual([1, 2, 3], r)

        r = utils.convert_to_integer_list('1s,2df,3d/')
        self.assertEqual([1, 2, 3], r)

        r = utils.convert_to_integer_list([1, 3, 4])
        self.assertEqual([1, 3, 4], r)

        r = utils.convert_to_integer_list('1,11,10')
        self.assertEqual([1, 11, 10], r)

        r = utils.convert_to_integer_list('1,11,10,QA')
        self.assertEqual([1, 11, 10, 'QA'], r)

    def test_geocode(self):
        loc = utils.geocode('1600 Pennsylvania Ave NW, Washington, DC 20500')

        self.assertEqual(round(loc['lat'], 3), 38.898)
        self.assertEqual(round(loc['lon'], 3), -77.037)
        self.assertRaises(ValueError, utils.geocode, 'Pennsylvania Ave NW, Washington, DC')
        self.assertEqual({'lat': 38.8987709, 'lon': -77.0351295},
                         utils.geocode('Pennsylvania Ave NW, Washington, DC', 10.))

    def test_convert_to_float_list(self):
        # correct input
        r = utils.convert_to_float_list('-1,2,-3')
        self.assertEqual([-1.0, 2.0, -3.0], r)

        # try other cobinations
        r = utils.convert_to_float_list('1, 2, 3')
        self.assertEqual([1.0, 2.0, 3.0], r)

        r = utils.convert_to_float_list('1s,2df,3d/')
        self.assertEqual([1.0, 2.0, 3.0], r)

        r = utils.convert_to_float_list([1, 3, 4])
        self.assertEqual([1, 3, 4], r)

        r = utils.convert_to_float_list('1,11,10')
        self.assertEqual([1.0, 11.0, 10.0], r)

    def test_adjust_bounding_box(self):

        # Test target bounds with origin bounds
        # should return the target bounds
        origin = (100, 10, 80, 20)
        target = (90, 15, 91, 15)

        self.assertEqual(utils.adjust_bounding_box(origin, target), target)

        # Test target bounds intersects with origin bounds
        # should return the expected bounds
        origin = (100, 10, 80, 20)
        target = (120, -5, 99, 15)
        expected = (100, 10, 99, 15)

        self.assertEqual(utils.adjust_bounding_box(origin, target), expected)

        # Test target bounds do not intersect with origin bounds at all
        # Target bounds are above origin bounds
        # should return the origin bounds
        origin = (100, 10, 80, 20)
        target = (120, -5, 110, 9)

        self.assertEqual(utils.adjust_bounding_box(origin, target), origin)

        # Target bounds are on the right side of origin bounds
        origin = (100, 10, 80, 20)
        target = (82, 23, 91, 26)

        self.assertEqual(utils.adjust_bounding_box(origin, target), origin)

        # Target bounds are below of origin bounds
        origin = (100, 10, 80, 20)
        target = (70, 11, 60, 18)

        self.assertEqual(utils.adjust_bounding_box(origin, target), origin)

        # Target bounds are on the left side of origin bounds
        origin = (100, 10, 80, 20)
        target = (80, -20, 79, -19)

        self.assertEqual(utils.adjust_bounding_box(origin, target), origin)

    def test_url_builder(self):

        self.assertEqual('http://example.com/segment1/segment2',
                         utils.url_builder(['/http://example.com', 'segment1/', '/segment2']))

        self.assertEqual('http://example.com/segment1/segment2',
                         utils.url_builder(('/http://example.com', 'segment1/', '/segment2',)))

        with self.assertRaises(AssertionError):
            utils.url_builder('example.com')

if __name__ == '__main__':
    unittest.main()



# Landsat Util
# License: CC0 1.0 Universal

"""Tests for downloader"""

import os
import errno
import shutil
import unittest
from tempfile import mkdtemp

import mock

from landsat.downloader import Downloader, RemoteFileDoesntExist, IncorrectSceneId
from landsat.settings import GOOGLE_STORAGE, S3_LANDSAT


class TestDownloader(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.temp_folder = mkdtemp()
        cls.d = Downloader(download_dir=cls.temp_folder)
        cls.scene = 'LT81360082013127LGN01'
        cls.scene_2 = 'LC82050312014229LGN00'
        cls.scene_s3 = 'LC80010092015051LGN00'
        cls.scene_s3_2 = 'LC82050312015136LGN00'
        cls.scene_size = 59239149

    @classmethod
    def tearDownClass(cls):
        try:
            shutil.rmtree(cls.temp_folder)
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise

    def assertSize(self, url, path):
        remote_size = self.d.get_remote_file_size(url)
        download_size = os.path.getsize(path)

        self.assertEqual(remote_size, download_size)

    @mock.patch('landsat.downloader.fetch')
    def test_download(self, mock_fetch):
        mock_fetch.return_value = True

        # download one scene
        self.d.download([self.scene])
        paths = self.d.download([self.scene])
        self.assertTrue(isinstance(paths, list))
        self.assertEqual([self.temp_folder + '/' + self.scene + '.tar.bz'],
                         paths)

        # download multiple scenes
        paths = self.d.download([self.scene, self.scene_2])
        test_paths = [self.temp_folder + '/' + self.scene + '.tar.bz',
                      self.temp_folder + '/' + self.scene_2 + '.tar.bz']
        self.assertTrue(isinstance(paths, list))
        self.assertEqual(test_paths, paths)

        # Test if error is raised when passing scene as string instead of list
        self.assertRaises(Exception, self.d.download, self.scene)

        # Test when passing band list along with sceneID
        paths = self.d.download([self.scene_s3, self.scene_s3_2], bands=[11])
        test_paths = [self.temp_folder + '/' + self.scene_s3,
                      self.temp_folder + '/' + self.scene_s3_2]
        self.assertEqual(test_paths, paths)

        # When passing scene as string, google storage download should be triggered
        paths = self.d.download([self.scene], bands=4)
        test_paths = [self.temp_folder + '/' + self.scene + '.tar.bz']
        self.assertEqual(test_paths, paths)

    @mock.patch('landsat.downloader.Downloader.google_storage')
    def test_download_google_when_amazon_is_unavailable(self, fake_google):
        """ Test whether google or amazon are correctly selected based on input """

        fake_google.return_value = False

        # Test if google is used when an image from 2014 is passed even if bands are provided
        self.d.download([self.scene], bands=[432])
        fake_google.assert_called_with(self.scene, self.d.download_dir)

    @mock.patch('landsat.downloader.fetch')
    def test_download_amazon_when_available(self, mock_fetch):
        """ Test whether google or amazon are correctly selected based on input """

        mock_fetch.return_value = True

        # Test if amazon is used when an image from 2015 is passed with bands
        paths = self.d.download([self.scene_s3], bands=[4, 3, 2])
        test_paths = [self.temp_folder + '/' + self.scene_s3]
        self.assertEqual(test_paths, paths)

    @mock.patch('landsat.downloader.fetch')
    def test_fetch(self, mock_fetch):
        mock_fetch.return_value = True

        sat = self.d.scene_interpreter(self.scene)
        url = self.d.google_storage_url(sat)

        self.assertTrue(self.d.fetch(url, self.temp_folder))

    def test_remote_file_size(self):

        url = self.d.google_storage_url(self.d.scene_interpreter(self.scene))
        size = self.d.get_remote_file_size(url)

        self.assertAlmostEqual(self.scene_size, size)

    def test_google_storage_url(self):
        sat = self.d.scene_interpreter(self.scene)

        string = self.d.google_storage_url(sat)
        expect = os.path.join(GOOGLE_STORAGE, 'L8/136/008/LT81360082013127LGN01.tar.bz')
        self.assertEqual(expect, string)

    def test_amazon_s3_url(self):
        sat = self.d.scene_interpreter(self.scene)
        string = self.d.amazon_s3_url(sat, 11)
        expect = os.path.join(S3_LANDSAT, 'L8/136/008/LT81360082013127LGN01/LT81360082013127LGN01_B11.TIF')

        self.assertEqual(expect, string)

    def test_remote_file_exist(self):
        # Exists and should return None

        self.assertIsNone(self.d.remote_file_exists(os.path.join(S3_LANDSAT, 'L8/003/017/LC80030172015001L'
                                                    'GN00/LC80030172015001LGN00_B6.TIF')))

        # Doesn't exist and should raise errror
        with self.assertRaises(RemoteFileDoesntExist):
            self.d.remote_file_exists(
                os.path.join(
                    S3_LANDSAT,
                    'L8/003/017/LC80030172015001LGN00/LC80030172015001LGN00_B34.TIF'
                )
            )

        # Doesn't exist and should raise errror
        with self.assertRaises(RemoteFileDoesntExist):
            self.d.remote_file_exists(
                os.path.join(
                    GOOGLE_STORAGE,
                    'L8/003/017/LC80030172015001LGN00/LC80030172015001LGN00_B6.TIF'
                )
            )

        # Exist and shouldn't raise error
        self.assertIsNone(self.d.remote_file_exists(os.path.join(GOOGLE_STORAGE,
                                                    'L8/003/017/LC80030172015001LGN00.tar.bz')))

    def test_scene_interpreter(self):
        # Test with correct input
        scene = 'LC80030172015001LGN00'
        ouput = self.d.scene_interpreter(scene)
        self.assertEqual({'path': '003', 'row': '017', 'sat': 'L8', 'scene': scene}, ouput)

        # Test with incorrect input
        self.assertRaises(Exception, self.d.scene_interpreter, 'LC80030172015001LGN')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# Landsat Util
# License: CC0 1.0 Universal

from __future__ import print_function, division, absolute_import

import argparse
import textwrap
import json
from os.path import join

try:
    from urllib.request import URLError
except ImportError:
    from urllib2 import URLError

from datetime import datetime
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse
import pycurl
from boto.exception import NoAuthHandlerFound

from .downloader import Downloader, IncorrectSceneId, RemoteFileDoesntExist, USGSInventoryAccessMissing
from .search import Search
from .uploader import Uploader
from .utils import reformat_date, convert_to_integer_list, timer, exit, get_file, convert_to_float_list
from .mixins import VerbosityMixin
from .image import Simple, PanSharpen, FileDoesNotExist
from .ndvi import NDVIWithManualColorMap, NDVI
from .__init__ import __version__
from . import settings


DESCRIPTION = """Landsat-util is a command line utility that makes it easy to
search, download, and process Landsat imagery.

    Commands:
        Search:
            landsat.py search [-p --pathrow] [--lat] [--lon] [--address] [-l LIMIT] [-s START] [-e END] [-c CLOUD]
                              [-h]

            optional arguments:
                -p, --pathrow       Paths and Rows in order separated by comma. Use quotes "001,003".
                                    Example: path,row,path,row 001,001,190,204

                --lat               Latitude

                --lon               Longitude

                --address           Street address

                -l LIMIT, --limit LIMIT
                                    Search return results limit default is 10

                -s START, --start START
                                    Start Date - Most formats are accepted e.g.
                                    Jun 12 2014 OR 06/12/2014

                -e END, --end END   End Date - Most formats are accepted e.g.
                                    Jun 12 2014 OR 06/12/2014

                --latest N          Returns the N latest images within the last 365 days.

                -c CLOUD, --cloud CLOUD
                                    Maximum cloud percentage. Default: 20 perct

                --json              Returns a bare JSON response

                --geojson           Returns a geojson response

                -h, --help          Show this help message and exit

        Download:
            landsat download sceneID [sceneID ...] [-h] [-b --bands]

            positional arguments:
                sceneID     Provide Full sceneIDs. You can add as many sceneIDs as you wish

                Example: landsat download LC81660392014196LGN00

            optional arguments:
                -b --bands          If you specify bands, landsat-util will try to download the band from S3.
                                    If the band does not exist, an error is returned

                -h, --help          Show this help message and exit

                -d, --dest          Destination path

                -p, --process       Process the image after download

                --pansharpen        Whether to also pansharpen the processed image.
                                    Pansharpening requires larger memory

                --ndvi              Calculates NDVI and produce a RGB GTiff with separate colorbar.

                --ndvigrey          Calculates NDVI and produce a greyscale GTiff.

                --clip              Clip the image with the bounding box provided. Values must be in WGS84 datum,
                                    and with longitude and latitude units of decimal degrees separated by comma.
                                    Example: --clip=-346.06658935546875,49.93531194616915,-345.4595947265625,
                                    50.2682767372753

                -u --upload         Upload to S3 after the image processing completed

                --key               Amazon S3 Access Key (You can also be set AWS_ACCESS_KEY_ID as
                                    Environment Variables)

                --secret            Amazon S3 Secret Key (You can also be set AWS_SECRET_ACCESS_KEY as
                                    Environment Variables)

                --bucket            Bucket name (required if uploading to s3)

                --region            URL to S3 region e.g. s3-us-west-2.amazonaws.com

                --force-unzip       Force unzip tar file

                --username          USGS Eros account Username (only works if the account has special
                                    inventory access). Username and password as a fallback if the image
                                    is not found on AWS S3 or Google Storage

                --password          USGS Eros account Password

        Process:
            landsat.py process path [-h] [-b --bands] [-p --pansharpen]

            positional arguments:
                path          Path to the landsat image folder or zip file

            optional arguments:
                -b --bands             Specify bands. The bands should be written in sequence with no spaces
                                    Default: Natural colors (432)
                                    Example --bands 432

                --pansharpen        Whether to also pansharpen the process image.
                                    Pansharpening requires larger memory

                --ndvi              Calculates NDVI and produce a RGB GTiff with separate colorbar.

                --ndvigrey          Calculates NDVI and produce a greyscale GTiff.

                --clip              Clip the image with the bounding box provided. Values must be in WGS84 datum,
                                    and with longitude and latitude units of decimal degrees separated by comma.
                                    Example: --clip=-346.06658935546875,49.93531194616915,-345.4595947265625,
                                    50.2682767372753

                -v, --verbose       Show verbose output

                -h, --help          Show this help message and exit

                -u --upload         Upload to S3 after the image processing completed

                --key               Amazon S3 Access Key (You can also be set AWS_ACCESS_KEY_ID as
                                    Environment Variables)

                --secret            Amazon S3 Secret Key (You can also be set AWS_SECRET_ACCESS_KEY as
                                    Environment Variables)

                --bucket            Bucket name (required if uploading to s3)

                --region            URL to S3 region e.g. s3-us-west-2.amazonaws.com

                --force-unzip       Force unzip tar file
"""


def args_options():
    """ Generates an arugment parser.

    :returns:
        Parser object
    """

    parser = argparse.ArgumentParser(prog='landsat',
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=textwrap.dedent(DESCRIPTION))

    subparsers = parser.add_subparsers(help='Landsat Utility',
                                       dest='subs')

    parser.add_argument('--version', action='version', version='%(prog)s version ' + __version__)

    # Search Logic
    parser_search = subparsers.add_parser('search',
                                          help='Search Landsat metadata')

    # Global search options
    parser_search.add_argument('-l', '--limit', default=10, type=int,
                               help='Search return results limit\n'
                               'default is 10')
    parser_search.add_argument('-s', '--start',
                               help='Start Date - Most formats are accepted '
                               'e.g. Jun 12 2014 OR 06/12/2014')
    parser_search.add_argument('-e', '--end',
                               help='End Date - Most formats are accepted '
                               'e.g. Jun 12 2014 OR 06/12/2014')
    parser_search.add_argument('--latest', default=-1, type=int,
                               help='returns the N latest images within the last 365 days')
    parser_search.add_argument('-c', '--cloud', type=float, default=100.0,
                               help='Maximum cloud percentage '
                               'default is 100 perct')
    parser_search.add_argument('-p', '--pathrow',
                               help='Paths and Rows in order separated by comma. Use quotes ("001").'
                               'Example: path,row,path,row 001,001,190,204')
    parser_search.add_argument('--lat', type=float, help='The latitude')
    parser_search.add_argument('--lon', type=float, help='The longitude')
    parser_search.add_argument('--address', type=str, help='The address')
    parser_search.add_argument('--json', action='store_true', help='Returns a bare JSON response')
    parser_search.add_argument('--geojson', action='store_true', help='Returns a geojson response')

    parser_download = subparsers.add_parser('download',
                                            help='Download images from Google Storage')
    parser_download.add_argument('scenes',
                                 metavar='sceneID',
                                 nargs="+",
                                 help="Provide Full sceneID, e.g. LC81660392014196LGN00")

    parser_download.add_argument('-b', '--bands', help='If you specify bands, landsat-util will try to download '
                                 'the band from S3. If the band does not exist, an error is returned', default=None)
    parser_download.add_argument('-d', '--dest', help='Destination path')
    parser_download.add_argument('-p', '--process', help='Process the image after download', action='store_true')
    parser_download.add_argument('--pansharpen', action='store_true',
                                 help='Whether to also pansharpen the process '
                                 'image. Pansharpening requires larger memory')
    parser_download.add_argument('--ndvi', action='store_true',
                                 help='Whether to run the NDVI process. If used, bands parameter is disregarded')
    parser_download.add_argument('--ndvigrey', action='store_true', help='Create an NDVI map in grayscale (grey)')
    parser_download.add_argument('--clip', help='Clip the image with the bounding box provided. Values must be in ' +
                                 'WGS84 datum, and with longitude and latitude units of decimal degrees ' +
                                 'separated by comma.' +
                                 'Example: --clip=-346.06658935546875,49.93531194616915,-345.4595947265625,' +
                                 '50.2682767372753')
    parser_download.add_argument('-u', '--upload', action='store_true',
                                 help='Upload to S3 after the image processing completed')
    parser_download.add_argument('--username', help='USGS Eros account Username (only works if the account has' +
                                 ' special inventory access). Username and password as a fallback if the image' +
                                 'is not found on AWS S3 or Google Storage')
    parser_download.add_argument('--password', help='USGS Eros username, used as a fallback')
    parser_download.add_argument('--key', help='Amazon S3 Access Key (You can also be set AWS_ACCESS_KEY_ID as '
                                 'Environment Variables)')
    parser_download.add_argument('--secret', help='Amazon S3 Secret Key (You can also be set AWS_SECRET_ACCESS_KEY '
                                 'as Environment Variables)')
    parser_download.add_argument('--bucket', help='Bucket name (required if uploading to s3)')
    parser_download.add_argument('--region', help='URL to S3 region e.g. s3-us-west-2.amazonaws.com')
    parser_download.add_argument('--force-unzip', help='Force unzip tar file', action='store_true')

    parser_process = subparsers.add_parser('process', help='Process Landsat imagery')
    parser_process.add_argument('path',
                                help='Path to the compressed image file')
    parser_process.add_argument('--pansharpen', action='store_true',
                                help='Whether to also pansharpen the process '
                                'image. Pansharpening requires larger memory')
    parser_process.add_argument('--ndvi', action='store_true', help='Create an NDVI map in color.')
    parser_process.add_argument('--ndvigrey', action='store_true', help='Create an NDVI map in grayscale (grey)')
    parser_process.add_argument('--clip', help='Clip the image with the bounding box provided. Values must be in ' +
                                'WGS84 datum, and with longitude and latitude units of decimal degrees ' +
                                'separated by comma.' +
                                'Example: --clip=-346.06658935546875,49.93531194616915,-345.4595947265625,' +
                                '50.2682767372753')
    parser_process.add_argument('-b', '--bands', help='specify band combinations. Default is 432'
                                'Example: --bands 321', default='432')
    parser_process.add_argument('-v', '--verbose', action='store_true',
                                help='Turn on verbosity')
    parser_process.add_argument('-u', '--upload', action='store_true',
                                help='Upload to S3 after the image processing completed')
    parser_process.add_argument('--key', help='Amazon S3 Access Key (You can also be set AWS_ACCESS_KEY_ID as '
                                'Environment Variables)')
    parser_process.add_argument('--secret', help='Amazon S3 Secret Key (You can also be set AWS_SECRET_ACCESS_KEY '
                                'as Environment Variables)')
    parser_process.add_argument('--bucket', help='Bucket name (required if uploading to s3)')
    parser_process.add_argument('--region', help='URL to S3 region e.g. s3-us-west-2.amazonaws.com')
    parser_process.add_argument('--force-unzip', help='Force unzip tar file', action='store_true')

    return parser


def main(args):
    """
    Main function - launches the program.

    :param args:
        The Parser arguments
    :type args:
        Parser object

    :returns:
        List

    :example:
        >>> ["The latitude and longitude values must be valid numbers", 1]
    """

    v = VerbosityMixin()

    if args:

        if 'clip' in args:
            bounds = convert_to_float_list(args.clip)
        else:
            bounds = None

        if args.subs == 'process':
            verbose = True if args.verbose else False
            force_unzip = True if args.force_unzip else False
            stored = process_image(args.path, args.bands, verbose, args.pansharpen, args.ndvi, force_unzip,
                                   args.ndvigrey, bounds)

            if args.upload:
                u = Uploader(args.key, args.secret, args.region)
                u.run(args.bucket, get_file(stored), stored)

            return ["The output is stored at %s" % stored]

        elif args.subs == 'search':

            try:
                if args.start:
                    args.start = reformat_date(parse(args.start))
                if args.end:
                    args.end = reformat_date(parse(args.end))
                if args.latest > 0:
                    args.limit = 25
                    end = datetime.now()
                    start = end - relativedelta(days=+365)
                    args.end = end.strftime("%Y-%m-%d")
                    args.start = start.strftime("%Y-%m-%d")
            except (TypeError, ValueError):
                return ["Your date format is incorrect. Please try again!", 1]

            s = Search()

            try:
                if args.lat is not None:
                    lat = float(args.lat)
                else:
                    lat = None

                if args.lon is not None:
                    lon = float(args.lon)
                else:
                    lon = None
            except ValueError:
                return ["The latitude and longitude values must be valid numbers", 1]

            address = args.address
            if address and (lat and lon):
                return ["Cannot specify both address and latitude-longitude"]

            result = s.search(paths_rows=args.pathrow,
                              lat=lat,
                              lon=lon,
                              address=address,
                              limit=args.limit,
                              start_date=args.start,
                              end_date=args.end,
                              cloud_max=args.cloud,
                              geojson=args.geojson)

            if 'status' in result:

                if result['status'] == 'SUCCESS':
                    if args.json:
                        return json.dumps(result)

                    if args.latest > 0:
                        datelist = []
                        for i in range(0, result['total_returned']):
                            datelist.append((result['results'][i]['date'], result['results'][i]))

                        datelist.sort(key=lambda tup: tup[0], reverse=True)
                        datelist = datelist[:args.latest]

                        result['results'] = []
                        for i in range(0, len(datelist)):
                            result['results'].append(datelist[i][1])
                            result['total_returned'] = len(datelist)

                    else:
                        v.output('%s items were found' % result['total'], normal=True, arrow=True)

                    if result['total'] > 100:
                        return ['Over 100 results. Please narrow your search', 1]
                    else:
                        v.output(json.dumps(result, sort_keys=True, indent=4), normal=True, color='green')
                    return ['Search completed!']

                elif result['status'] == 'error':
                    return [result['message'], 1]

            if args.geojson:
                return json.dumps(result)

        elif args.subs == 'download':
            d = Downloader(download_dir=args.dest, usgs_user=args.username, usgs_pass=args.password)
            try:
                bands = convert_to_integer_list(args.bands)

                if args.process:
                    if args.pansharpen:
                        bands.append(8)

                    if args.ndvi or args.ndvigrey:
                        bands = [4, 5]

                    if not args.bands:
                        bands = [4, 3, 2]

                files = d.download(args.scenes, bands)

                if args.process:
                    if not args.bands:
                        args.bands = '432'
                    force_unzip = True if args.force_unzip else False
                    for f in files:
                        stored = process_image(f, args.bands, False, args.pansharpen, args.ndvi, force_unzip,
                                               args.ndvigrey, bounds=bounds)

                        if args.upload:
                            try:
                                u = Uploader(args.key, args.secret, args.region)
                            except NoAuthHandlerFound:
                                return ["Could not authenticate with AWS", 1]
                            except URLError:
                                return ["Connection timeout. Probably the region parameter is incorrect", 1]
                            u.run(args.bucket, get_file(stored), stored)

                    return ['The output is stored at %s' % stored, 0]
                else:
                    return ['Download Completed', 0]
            except IncorrectSceneId:
                return ['The SceneID provided was incorrect', 1]
            except (RemoteFileDoesntExist, USGSInventoryAccessMissing) as e:
                return [e.message, 1]


def process_image(path, bands=None, verbose=False, pansharpen=False, ndvi=False, force_unzip=None,
                  ndvigrey=False, bounds=None):
    """ Handles constructing and image process.

    :param path:
        The path to the image that has to be processed
    :type path:
        String
    :param bands:
        List of bands that has to be processed. (optional)
    :type bands:
        List
    :param verbose:
        Sets the level of verbosity. Default is False.
    :type verbose:
        boolean
    :param pansharpen:
        Whether to pansharpen the image. Default is False.
    :type pansharpen:
        boolean

    :returns:
        (String) path to the processed image
    """
    try:
        bands = convert_to_integer_list(bands)
        if pansharpen:
            p = PanSharpen(path, bands=bands, dst_path=settings.PROCESSED_IMAGE,
                           verbose=verbose, force_unzip=force_unzip, bounds=bounds)
        elif ndvigrey:
            p = NDVI(path, verbose=verbose, dst_path=settings.PROCESSED_IMAGE, force_unzip=force_unzip, bounds=bounds)
        elif ndvi:
            p = NDVIWithManualColorMap(path, dst_path=settings.PROCESSED_IMAGE,
                                       verbose=verbose, force_unzip=force_unzip, bounds=bounds)
        else:
            p = Simple(path, bands=bands, dst_path=settings.PROCESSED_IMAGE, verbose=verbose, force_unzip=force_unzip,
                       bounds=bounds)

    except IOError as err:
        exit(str(err), 1)
    except FileDoesNotExist as err:
        exit(str(err), 1)

    return p.run()


def __main__():

    global parser
    parser = args_options()
    args = parser.parse_args()
    if args.subs == 'search' and (hasattr(args, 'json') or hasattr(args, 'geojson')):
            print(main(args))
    else:
        with timer():
            exit(*main(args))

if __name__ == "__main__":
    try:
        __main__()
    except (KeyboardInterrupt, pycurl.error):
        exit('Received Ctrl + C... Exiting! Bye.', 1)



# Landsat Util
# License: CC0 1.0 Universal

"""Tests for landsat"""

import json
import unittest
import subprocess
import errno
import shutil
from os.path import join

from jsonschema import validate
import mock

import landsat.landsat as landsat
from tests import geojson_schema


class TestLandsat(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.parser = landsat.args_options()
        cls.mock_path = 'path/to/folder'

    @classmethod
    def tearDownClass(cls):
        try:
            shutil.rmtree('path')
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise

    def test_incorrect_date(self):
        """ Test search with incorrect date input """

        args = ['search', '--start', 'berlin', '--end', 'january 10 2014']

        self.assertEquals(landsat.main(self.parser.parse_args(args)),
                          ['Your date format is incorrect. Please try again!', 1])

    def test_too_many_results(self):
        """ Test when search return too many results """

        args = ['search', '--cloud', '100', '-p', '205,022,206,022,204,022']

        self.assertEquals(landsat.main(self.parser.parse_args(args)),
                          ['Over 100 results. Please narrow your search', 1])

    def test_search_pr_correct(self):
        """Test Path Row search with correct input"""
        args = ['search', '--start', 'january 1 2013', '--end',
                'january 10 2014', '-p', '008,008']

        self.assertEquals(landsat.main(self.parser.parse_args(args)),
                          ['Search completed!'])

    def test_search_lat_lon(self):
        """Test Latitude Longitude search with correct input"""
        args = ['search', '--start', 'may 01 2013', '--end', 'may 08 2013',
                '--lat', '38.9107203', '--lon', '-77.0290116']

        self.assertEquals(landsat.main(self.parser.parse_args(args)),
                          ['Search completed!'])

    def test_search_pr_wrong_input(self):
        """Test Path Row search with incorrect input"""
        args = ['search', '-p', 'what?']

        self.assertEquals(landsat.main(self.parser.parse_args(args)), None)

    def test_search_json_output(self):
        """Test json output in search"""
        args = ['search', '--latest', '10', '--json']

        output = landsat.main(self.parser.parse_args(args))
        j = json.loads(output)

        self.assertEquals(type(j), dict)

    def test_search_geojson_output(self):
        """Test json output in search"""
        args = ['search', '--latest', '10', '--geojson']

        output = landsat.main(self.parser.parse_args(args))
        j = json.loads(output)

        self.assertIsNone(validate(j, geojson_schema))
        self.assertEquals(type(j), dict)

    @mock.patch('landsat.landsat.Downloader')
    def test_download_correct(self, mock_downloader):
        """Test download command with correct input"""
        mock_downloader.download.return_value = True

        args = ['download', 'LC80010092015051LGN00', '-b', '11,', '-d', self.mock_path]
        output = landsat.main(self.parser.parse_args(args))
        mock_downloader.assert_called_with(download_dir=self.mock_path, usgs_pass=None, usgs_user=None)
        mock_downloader.return_value.download.assert_called_with(['LC80010092015051LGN00'], [11])
        self.assertEquals(output, ['Download Completed', 0])

    @mock.patch('landsat.landsat.Downloader')
    def test_download_correct_zip(self, mock_downloader):
        """Download command should download zip if no bands are given"""
        mock_downloader.download.return_value = True

        args = ['download', 'LC80010092015051LGN00', '-d', self.mock_path]
        output = landsat.main(self.parser.parse_args(args))
        mock_downloader.assert_called_with(download_dir=self.mock_path, usgs_pass=None, usgs_user=None)
        mock_downloader.return_value.download.assert_called_with(['LC80010092015051LGN00'], None)
        self.assertEquals(output, ['Download Completed', 0])

    @mock.patch('landsat.landsat.process_image')
    @mock.patch('landsat.landsat.Downloader.download')
    def test_download_no_bands_with_process(self, mock_downloader, mock_process):
        """Download command should not download zip if no bands are given but process flag is used"""
        mock_downloader.return_value = {'LC80010092015051LGN00': 'aws'}
        mock_process.return_value = 'image.TIF'

        args = ['download', 'LC80010092015051LGN00', '-p', '-d', self.mock_path]
        output = landsat.main(self.parser.parse_args(args))
        mock_downloader.assert_called_with(['LC80010092015051LGN00'], [4, 3, 2])
        self.assertEquals(output, ["The output is stored at image.TIF", 0])

    def test_download_incorrect(self):
        """Test download command with incorrect input"""
        args = ['download', 'LT813600']

        self.assertEquals(landsat.main(self.parser.parse_args(args)),
                          ['The SceneID provided was incorrect', 1])

    @mock.patch('landsat.landsat.process_image')
    @mock.patch('landsat.downloader.fetch')
    def test_download_process_continuous(self, fetch, mock_process):
        """Test download and process commands together"""
        fetch.return_value = True
        mock_process.return_value = 'image.TIF'

        args = ['download', 'LC80010092015051LGN00', 'LC80470222014354LGN00', '-b', '432', '-d', self.mock_path, '-p']
        output = landsat.main(self.parser.parse_args(args))
        mock_process.assert_called_with('path/to/folder/LC80470222014354LGN00', '432',
                                        False, False, False, False, False, bounds=None)
        self.assertEquals(output, ["The output is stored at image.TIF", 0])

        # Call with force unzip flag
        args = ['download', 'LC80010092015051LGN00', 'LC80470222014354LGN00', '-b', '432', '-d',
                self.mock_path, '-p', '--force-unzip']
        output = landsat.main(self.parser.parse_args(args))
        mock_process.assert_called_with('path/to/folder/LC80470222014354LGN00', '432', False, False, False,
                                        True, False, bounds=None)
        self.assertEquals(output, ["The output is stored at image.TIF", 0])

        # Call with pansharpen
        args = ['download', 'LC80010092015051LGN00', 'LC80470222014354LGN00', '-b', '432', '-d',
                self.mock_path, '-p', '--pansharpen']
        output = landsat.main(self.parser.parse_args(args))
        mock_process.assert_called_with('path/to/folder/LC80470222014354LGN00', '432', False, True, False,
                                        False, False, bounds=None)
        self.assertEquals(output, ["The output is stored at image.TIF", 0])

        # Call with pansharpen and clipping
        args = ['download', 'LC80010092015051LGN00', 'LC80470222014354LGN00', '-b', '432', '-d',
                self.mock_path, '-p', '--pansharpen', '--clip', '"-180,-180,0,0"']
        output = landsat.main(self.parser.parse_args(args))
        mock_process.assert_called_with('path/to/folder/LC80470222014354LGN00', '432', False, True, False,
                                        False, False, bounds=[-180.0, -180.0, 0.0, 0.0])
        self.assertEquals(output, ["The output is stored at image.TIF", 0])

        # Call with ndvi
        args = ['download', 'LC80010092015051LGN00', 'LC80470222014354LGN00', '-b', '432', '-d',
                self.mock_path, '-p', '--ndvi']
        output = landsat.main(self.parser.parse_args(args))
        mock_process.assert_called_with('path/to/folder/LC80470222014354LGN00', '432', False, False, True,
                                        False, False, bounds=None)
        self.assertEquals(output, ["The output is stored at image.TIF", 0])

        # Call with ndvigrey
        args = ['download', 'LC80010092015051LGN00', 'LC80470222014354LGN00', '-b', '432', '-d',
                self.mock_path, '-p', '--ndvigrey']
        output = landsat.main(self.parser.parse_args(args))
        mock_process.assert_called_with('path/to/folder/LC80470222014354LGN00', '432', False, False, False,
                                        False, True, bounds=None)
        self.assertEquals(output, ["The output is stored at image.TIF", 0])

    @mock.patch('landsat.landsat.Uploader')
    @mock.patch('landsat.landsat.process_image')
    @mock.patch('landsat.downloader.fetch')
    def test_download_process_continuous_with_upload(self, fetch, mock_process, mock_upload):
        """Test download and process commands together"""
        fetch.return_value = True
        mock_process.return_value = 'image.TIF'
        mock_upload.run.return_value = True

        args = ['download', 'LC80010092015051LGN00', '-b', '432', '-d', self.mock_path, '-p',
                '-u', '--key', 'somekey', '--secret', 'somesecret', '--bucket', 'mybucket', '--region', 'this']
        output = landsat.main(self.parser.parse_args(args))
        # mock_downloader.assert_called_with(['LC80010092015051LGN00'], [4, 3, 2])
        mock_process.assert_called_with('path/to/folder/LC80010092015051LGN00', '432', False, False, False,
                                        False, False, bounds=None)
        mock_upload.assert_called_with('somekey', 'somesecret', 'this')
        mock_upload.return_value.run.assert_called_with('mybucket', 'image.TIF', 'image.TIF')
        self.assertEquals(output, ['The output is stored at image.TIF', 0])

    @mock.patch('landsat.landsat.process_image')
    @mock.patch('landsat.downloader.fetch')
    def test_download_process_continuous_with_wrong_args(self, fetch, mock_process):
        """Test download and process commands together"""
        fetch.return_value = True
        mock_process.return_value = 'image.TIF'

        args = ['download', 'LC80010092015051LGN00', '-b', '432', '-d', self.mock_path, '-p',
                '-u', '--region', 'whatever']
        output = landsat.main(self.parser.parse_args(args))
        mock_process.assert_called_with('path/to/folder/LC80010092015051LGN00', '432', False, False, False,
                                        False, False, bounds=None)
        self.assertEquals(output, ['Could not authenticate with AWS', 1])

    @mock.patch('landsat.landsat.process_image')
    def test_process_correct(self, mock_process):
        """Test process command with correct input"""
        mock_process.return_value = 'image.TIF'

        args = ['process', 'path/to/folder/LC80010092015051LGN00']
        output = landsat.main(self.parser.parse_args(args))

        mock_process.assert_called_with('path/to/folder/LC80010092015051LGN00', '432',
                                        False, False, False, False, False, None)
        self.assertEquals(output, ["The output is stored at image.TIF"])

    @mock.patch('landsat.landsat.process_image')
    def test_process_correct_with_clipping(self, mock_process):
        """Test process command with correct input"""
        mock_process.return_value = 'image.TIF'

        args = ['process', 'path/to/folder/LC80010092015051LGN00', '--clip', '"-180,-180,0,0"']
        output = landsat.main(self.parser.parse_args(args))

        mock_process.assert_called_with('path/to/folder/LC80010092015051LGN00', '432',
                                        False, False, False, False, False, [-180.0, -180.0, 0.0, 0.0])
        self.assertEquals(output, ["The output is stored at image.TIF"])

    @mock.patch('landsat.landsat.process_image')
    def test_process_correct_pansharpen(self, mock_process):
        """Test process command with correct input and pansharpening"""
        mock_process.return_value = 'image.TIF'

        args = ['process', '--pansharpen', 'path/to/folder/LC80010092015051LGN00']
        output = landsat.main(self.parser.parse_args(args))

        mock_process.assert_called_with('path/to/folder/LC80010092015051LGN00', '432', False, True, False, False,
                                        False, None)
        self.assertEquals(output, ["The output is stored at image.TIF"])

    @mock.patch('landsat.landsat.process_image')
    def test_process_correct_ndvi(self, mock_process):
        """Test process command with correct input and ndvi"""
        mock_process.return_value = 'image.TIF'

        args = ['process', '--ndvi', 'path/to/folder/LC80010092015051LGN00']
        output = landsat.main(self.parser.parse_args(args))

        mock_process.assert_called_with('path/to/folder/LC80010092015051LGN00', '432', False, False, True, False,
                                        False, None)
        self.assertEquals(output, ["The output is stored at image.TIF"])

    def test_process_incorrect(self):
        """Test process command with incorrect input"""
        args = ['process', 'whatever']

        try:
            landsat.main(self.parser.parse_args(args))
        except SystemExit as e:
            self.assertEqual(e.code, 1)

    def check_command_line(self):
        """ Check if the commandline performs correctly """

        self.assertEqual(subprocess.call(['python', join(self.base_dir, '../landsat.py'), '-h']), 0)


if __name__ == '__main__':
    unittest.main()



# Landsat Util
# License: CC0 1.0 Universal

from __future__ import print_function, division, absolute_import

from xml.etree import ElementTree
from os.path import join, exists, getsize

import requests
from usgs import api, USGSError
from homura import download as fetch

from .utils import check_create_folder, url_builder
from .mixins import VerbosityMixin
from . import settings


class RemoteFileDoesntExist(Exception):
    """ Exception to be used when the remote file does not exist """
    pass


class IncorrectSceneId(Exception):
    """ Exception to be used when scene id is incorrect """
    pass


class USGSInventoryAccessMissing(Exception):
    """ Exception for when User does not have Inventory Service access """
    pass


class Downloader(VerbosityMixin):
    """ The downloader class """

    def __init__(self, verbose=False, download_dir=None, usgs_user=None, usgs_pass=None):
        self.download_dir = download_dir if download_dir else settings.DOWNLOAD_DIR
        self.google = settings.GOOGLE_STORAGE
        self.s3 = settings.S3_LANDSAT
        self.usgs_user = usgs_user
        self.usgs_pass = usgs_pass

        # Make sure download directory exist
        check_create_folder(self.download_dir)

    def download(self, scenes, bands=None):
        """
        Download scenese from Google Storage or Amazon S3 if bands are provided

        :param scenes:
            A list of scene IDs
        :type scenes:
            List
        :param bands:
            A list of bands. Default value is None.
        :type scenes:
            List

        :returns:
            (List) includes downloaded scenes as key and source as value (aws or google)
        """

        if isinstance(scenes, list):
            files = []

            for scene in scenes:

                # for all scenes if bands provided, first check AWS, if the bands exist
                # download them, otherwise use Google and then USGS.
                try:
                    # if bands are not provided, directly go to Goodle and then USGS
                    if not isinstance(bands, list):
                        raise RemoteFileDoesntExist
                    files.append(self.amazon_s3(scene, bands))

                except RemoteFileDoesntExist:
                    try:
                        files.append(self.google_storage(scene, self.download_dir))
                    except RemoteFileDoesntExist:
                        files.append(self.usgs_eros(scene, self.download_dir))

            return files

        else:
            raise Exception('Expected sceneIDs list')

    def usgs_eros(self, scene, path):
        """ Downloads the image from USGS """

        # download from usgs if login information is provided
        if self.usgs_user and self.usgs_pass:
            try:
                api_key = api.login(self.usgs_user, self.usgs_pass)
            except USGSError as e:
                error_tree = ElementTree.fromstring(str(e.message))
                error_text = error_tree.find("SOAP-ENV:Body/SOAP-ENV:Fault/faultstring", api.NAMESPACES).text
                raise USGSInventoryAccessMissing(error_text)

            download_url = api.download('LANDSAT_8', 'EE', [scene], api_key=api_key)
            if download_url:
                self.output('Source: USGS EarthExplorer', normal=True, arrow=True)
                return self.fetch(download_url[0], path)

            raise RemoteFileDoesntExist('%s is not available on AWS S3, Google or USGS Earth Explorer' % scene)
        raise RemoteFileDoesntExist('%s is not available on AWS S3 or Google Storage' % scene)

    def google_storage(self, scene, path):
        """
        Google Storage Downloader.

        :param scene:
            The scene id
        :type scene:
            String
        :param path:
            The directory path to where the image should be stored
        :type path:
            String

        :returns:
            Boolean
        """

        sat = self.scene_interpreter(scene)
        url = self.google_storage_url(sat)

        self.remote_file_exists(url)

        self.output('Source: Google Storage', normal=True, arrow=True)
        return self.fetch(url, path)

    def amazon_s3(self, scene, bands):
        """
        Amazon S3 downloader
        """

        sat = self.scene_interpreter(scene)

        # Always grab MTL.txt and QA band if bands are specified
        if 'BQA' not in bands:
            bands.append('QA')

        if 'MTL' not in bands:
            bands.append('MTL')

        urls = []

        for band in bands:
            # get url for the band
            url = self.amazon_s3_url(sat, band)

            # make sure it exist
            self.remote_file_exists(url)
            urls.append(url)

        # create folder
        path = check_create_folder(join(self.download_dir, scene))

        self.output('Source: AWS S3', normal=True, arrow=True)
        for url in urls:
            self.fetch(url, path)

        return path

    def fetch(self, url, path):
        """ Downloads the given url.

        :param url:
            The url to be downloaded.
        :type url:
            String
        :param path:
            The directory path to where the image should be stored
        :type path:
            String
        :param filename:
            The filename that has to be downloaded
        :type filename:
            String

        :returns:
            Boolean
        """

        segments = url.split('/')
        filename = segments[-1]

        # remove query parameters from the filename
        filename = filename.split('?')[0]

        self.output('Downloading: %s' % filename, normal=True, arrow=True)

        # print(join(path, filename))
        # raise Exception
        if exists(join(path, filename)):
            size = getsize(join(path, filename))
            if size == self.get_remote_file_size(url):
                self.output('%s already exists on your system' % filename, normal=True, color='green', indent=1)

        else:
            fetch(url, path)
        self.output('stored at %s' % path, normal=True, color='green', indent=1)

        return join(path, filename)

    def google_storage_url(self, sat):
        """
        Returns a google storage url the contains the scene provided.

        :param sat:
            Expects an object created by scene_interpreter method
        :type sat:
            dict

        :returns:
            (String) The URL to a google storage file
        """
        filename = sat['scene'] + '.tar.bz'
        return url_builder([self.google, sat['sat'], sat['path'], sat['row'], filename])

    def amazon_s3_url(self, sat, band):
        """
        Return an amazon s3 url the contains the scene and band provided.

        :param sat:
            Expects an object created by scene_interpreter method
        :type sat:
            dict
        :param filename:
            The filename that has to be downloaded from Amazon
        :type filename:
            String

        :returns:
            (String) The URL to a S3 file
        """
        if band != 'MTL':
            filename = '%s_B%s.TIF' % (sat['scene'], band)
        else:
            filename = '%s_%s.txt' % (sat['scene'], band)

        return url_builder([self.s3, sat['sat'], sat['path'], sat['row'], sat['scene'], filename])

    def remote_file_exists(self, url):
        """ Checks whether the remote file exists.

        :param url:
            The url that has to be checked.
        :type url:
            String

        :returns:
            **True** if remote file exists and **False** if it doesn't exist.
        """
        status = requests.head(url).status_code

        if status != 200:
            raise RemoteFileDoesntExist

    def get_remote_file_size(self, url):
        """ Gets the filesize of a remote file.

        :param url:
            The url that has to be checked.
        :type url:
            String

        :returns:
            int
        """
        headers = requests.head(url).headers
        return int(headers['content-length'])

    def scene_interpreter(self, scene):
        """ Conver sceneID to rows, paths and dates.

        :param scene:
            The scene ID.
        :type scene:
            String

        :returns:
            dict

        :Example output:

        >>> anatomy = {
                'path': None,
                'row': None,
                'sat': None,
                'scene': scene
            }
        """
        anatomy = {
            'path': None,
            'row': None,
            'sat': None,
            'scene': scene
        }
        if isinstance(scene, str) and len(scene) == 21:
            anatomy['path'] = scene[3:6]
            anatomy['row'] = scene[6:9]
            anatomy['sat'] = 'L' + scene[2:3]

            return anatomy
        else:
            raise IncorrectSceneId('Received incorrect scene')


if __name__ == '__main__':

    d = Downloader()

    # d.download(['LC81990242015046LGN00', 'LC80030172015001LGN00'])



state = {}


class MockBotoS3MultipartUpload():
    def __init__(self):
        self.data = state['mock_boto_s3_multipart_upload_data']

    def upload_part_from_file(self, f, part_no, cb=None, num_cb=None):
        self.data.append(f.read())

    def complete_upload(self):
        pass

    def cancel_upload(self):
        pass


class MockBotoS3Bucket():
    def lookup(self, key):
        pass

    def initiate_multipart_upload(self, key):
        return MockBotoS3MultipartUpload()


class S3Connection():
    def __init__(self, key, secret, is_secure=None, host=None):
        pass

    def get_bucket(self, bucket_name):
        return MockBotoS3Bucket()



# Landsat Util
# License: CC0 1.0 Universal

from __future__ import print_function, division, absolute_import

import json
import time
import requests

from . import settings
from .utils import three_digit, create_paired_list, geocode


class Search(object):
    """ The search class """

    def __init__(self):
        self.api_url = settings.API_URL

    def search(self, paths_rows=None, lat=None, lon=None, address=None, start_date=None, end_date=None, cloud_min=None,
               cloud_max=None, limit=1, geojson=False):
        """
        The main method of Search class. It searches Development Seed's Landsat API.

        :param paths_rows:
            A string in this format: "003,003,004,004". Must be in pairs and separated by comma.
        :type paths_rows:
            String
        :param lat:
            The latitude
        :type lat:
            String, float, integer
        :param lon:
            The The longitude
        :type lon:
            String, float, integer
        :param address:
            The address
        :type address:
            String
        :param start_date:
            Date string. format: YYYY-MM-DD
        :type start_date:
            String
        :param end_date:
            date string. format: YYYY-MM-DD
        :type end_date:
            String
        :param cloud_min:
            float specifying the minimum percentage. e.g. 4.3
        :type cloud_min:
            float
        :param cloud_max:
            float specifying the maximum percentage. e.g. 78.9
        :type cloud_max:
            float
        :param limit:
            integer specigying the maximum results return.
        :type limit:
            integer
        :param geojson:
            boolean specifying whether to return a geojson object
        :type geojson:
            boolean

        :returns:
            dict

        :example:
            >>> search = Search()
            >>> search('003,003', '2014-01-01', '2014-06-01')
            >>> {
                    'status': u'SUCCESS',
                    'total_returned': 1,
                    'total': 1,
                    'limit': 1
                    'results': [
                        {
                            'sat_type': u'L8',
                            'sceneID': u'LC80030032014142LGN00',
                            'date': u'2014-05-22',
                            'path': u'003',
                            'thumbnail': u'http://....../landsat_8/2014/003/003/LC80030032014142LGN00.jpg',
                            'cloud': 33.36,
                            'row': u'003
                        }
                    ]
                }
        """

        search_string = self.query_builder(paths_rows, lat, lon, address, start_date, end_date, cloud_min, cloud_max)

        # Have to manually build the URI to bypass requests URI encoding
        # The api server doesn't accept encoded URIs

        r = requests.get('%s?search=%s&limit=%s' % (self.api_url, search_string, limit))

        r_dict = json.loads(r.text)
        result = {}

        if 'error' in r_dict:
            result['status'] = u'error'
            result['code'] = r_dict['error']['code']
            result['message'] = r_dict['error']['message']

        elif 'meta' in r_dict:
            if geojson:
                result = {
                    'type': 'FeatureCollection',
                    'features': []
                }
                for r in r_dict['results']:
                    feature = {
                        'type': 'Feature',
                        'properties': {
                            'sceneID': r['sceneID'],
                            'row': three_digit(r['row']),
                            'path': three_digit(r['path']),
                            'thumbnail': r['browseURL'],
                            'date': r['acquisitionDate'],
                            'cloud': r['cloudCoverFull']
                        },
                        'geometry': {
                            'type': 'Polygon',
                            'coordinates': [
                                [
                                    [r['upperLeftCornerLongitude'], r['upperLeftCornerLatitude']],
                                    [r['lowerLeftCornerLongitude'], r['lowerLeftCornerLatitude']],
                                    [r['lowerRightCornerLongitude'], r['lowerRightCornerLatitude']],
                                    [r['upperRightCornerLongitude'], r['upperRightCornerLatitude']],
                                    [r['upperLeftCornerLongitude'], r['upperLeftCornerLatitude']]
                                ]
                            ]
                        }
                    }

                    result['features'].append(feature)

            else:
                result['status'] = u'SUCCESS'
                result['total'] = r_dict['meta']['found']
                result['limit'] = r_dict['meta']['limit']
                result['total_returned'] = len(r_dict['results'])
                result['results'] = [{'sceneID': i['sceneID'],
                                      'sat_type': u'L8',
                                      'path': three_digit(i['path']),
                                      'row': three_digit(i['row']),
                                      'thumbnail': i['browseURL'],
                                      'date': i['acquisitionDate'],
                                      'cloud': i['cloudCoverFull']}
                                     for i in r_dict['results']]

        return result

    def query_builder(self, paths_rows=None, lat=None, lon=None, address=None, start_date=None, end_date=None,
                      cloud_min=None, cloud_max=None):
        """ Builds the proper search syntax (query) for Landsat API.

        :param paths_rows:
            A string in this format: "003,003,004,004". Must be in pairs and separated by comma.
        :type paths_rows:
            String
        :param lat:
            The latitude
        :type lat:
            String, float, integer
        :param lon:
            The The longitude
        :type lon:
            String, float, integer
        :param address:
            The address
        :type address:
            String
        :param start_date:
            Date string. format: YYYY-MM-DD
        :type start_date:
            String
        :param end_date:
            date string. format: YYYY-MM-DD
        :type end_date:
            String
        :param cloud_min:
            float specifying the minimum percentage. e.g. 4.3
        :type cloud_min:
            float
        :param cloud_max:
            float specifying the maximum percentage. e.g. 78.9
        :type cloud_max:
            float

        :returns:
            String
        """

        query = []
        or_string = ''
        and_string = ''
        search_string = ''

        if paths_rows:
            # Coverting rows and paths to paired list
            new_array = create_paired_list(paths_rows)
            paths_rows = ['(%s)' % self.row_path_builder(i[0], i[1]) for i in new_array]
            or_string = '+OR+'.join(map(str, paths_rows))

        if start_date and end_date:
            query.append(self.date_range_builder(start_date, end_date))
        elif start_date:
            query.append(self.date_range_builder(start_date, '2100-01-01'))
        elif end_date:
            query.append(self.date_range_builder('2009-01-01', end_date))

        if cloud_min and cloud_max:
            query.append(self.cloud_cover_prct_range_builder(cloud_min, cloud_max))
        elif cloud_min:
            query.append(self.cloud_cover_prct_range_builder(cloud_min, '100'))
        elif cloud_max:
            query.append(self.cloud_cover_prct_range_builder('-1', cloud_max))

        if address:
            query.append(self.address_builder(address))
        elif (lat is not None) and (lon is not None):
            query.append(self.lat_lon_builder(lat, lon))

        if query:
            and_string = '+AND+'.join(map(str, query))

        if and_string and or_string:
            search_string = and_string + '+AND+(' + or_string + ')'
        else:
            search_string = or_string + and_string

        return search_string

    def row_path_builder(self, path='', row=''):
        """
        Builds row and path query.

        :param path:
            Landsat path. Must be three digits
        :type path:
            String
        :param row:
            Landsat row. Must be three digits
        :type row:
            String

        :returns:
            String
        """
        return 'path:%s+AND+row:%s' % (path, row)

    def date_range_builder(self, start='2013-02-11', end=None):
        """
        Builds date range query.

        :param start:
            Date string. format: YYYY-MM-DD
        :type start:
            String
        :param end:
            date string. format: YYYY-MM-DD
        :type end:
            String

        :returns:
            String
        """
        if not end:
            end = time.strftime('%Y-%m-%d')

        return 'acquisitionDate:[%s+TO+%s]' % (start, end)

    def cloud_cover_prct_range_builder(self, min=0, max=100):
        """
        Builds cloud cover percentage range query.

        :param min:
            float specifying the minimum percentage. Default is 0
        :type min:
            float
        :param max:
            float specifying the maximum percentage. Default is 100
        :type max:
            float

        :returns:
            String
        """
        return 'cloudCoverFull:[%s+TO+%s]' % (min, max)

    def address_builder(self, address):
        """ Builds lat and lon query from a geocoded address.

        :param address:
            The address
        :type address:
            String

        :returns:
            String
        """
        geocoded = geocode(address)
        return self.lat_lon_builder(**geocoded)

    def lat_lon_builder(self, lat=0, lon=0):
        """ Builds lat and lon query.

        :param lat:
            The latitude. Default is 0
        :type lat:
            float
        :param lon:
            The The longitude. Default is 0
        :type lon:
            float

        :returns:
            String
        """
        return ('upperLeftCornerLatitude:[%s+TO+1000]+AND+lowerRightCornerLatitude:[-1000+TO+%s]'
                '+AND+lowerLeftCornerLongitude:[-1000+TO+%s]+AND+upperRightCornerLongitude:[%s+TO+1000]'
                % (lat, lat, lon, lon))



# Landsat Util
# License: CC0 1.0 Universal

# Some of the tests are from pys3upload (https://github.com/leetreveil/pys3upload)

"""Tests for uploader"""

import os
import sys
import unittest
import threading

import mock

from landsat.uploader import Uploader, upload, upload_part, data_collector
from .mocks import S3Connection, state


class TestUploader(unittest.TestCase):

    @mock.patch('landsat.uploader.S3Connection', S3Connection)
    def test_upload_to_s3(self):
        state['mock_boto_s3_multipart_upload_data'] = []
        base_dir = os.path.abspath(os.path.dirname(__file__))
        landsat_image = os.path.join(base_dir, 'samples/mock_upload')
        f = open(landsat_image, 'rb').readlines()

        u = Uploader('some_key', 'some_secret')
        u.run('some bucket', 'mock_upload', landsat_image)

        self.assertEqual(state['mock_boto_s3_multipart_upload_data'], f)


class upload_tests(unittest.TestCase):

    def test_should_be_able_to_upload_data(self):
        input = [b'12', b'345']
        state['mock_boto_s3_multipart_upload_data'] = []
        conn = S3Connection('some_key', 'some_secret', True)
        upload('test_bucket', 'some_key', 'some_secret', input, 'some_key', connection=conn)
        self.assertEqual(state['mock_boto_s3_multipart_upload_data'], [b'12', b'345'])


class upload_part_tests(unittest.TestCase):

    def test_should_return_error_when_upload_func_raises_error(self):
        def upload_func(*args, **kwargs):
            raise Exception()

        with self.assertRaises(threading.ThreadError):
            raise upload_part(upload_func, '_', '_', '_')

    def test_should_retry_upload_five_times(self):
        counter = [0]

        def upload_func(*args, **kwargs):
            counter[0] += 1
            raise Exception()

        upload_part(upload_func, b'_', b'_', b'_')
        self.assertEqual(counter[0], 5)


class doc_collector_tests(unittest.TestCase):

    def test_should_be_able_to_read_every_byte_of_data(self):
        input = [b'12345']
        result = list(data_collector(input, def_buf_size=3))
        self.assertEqual(result, [b'123', b'45'])

    def test_should_be_able_to_read_single_yield(self):
        input = [b'123']
        result = list(data_collector(input, def_buf_size=3))
        self.assertEqual(result, [b'123'])

    def test_should_be_able_to_yield_data_less_than_buffer_size(self):
        input = [b'123']
        result = list(data_collector(input, def_buf_size=6))
        self.assertEqual(result, [b'123'])

    def test_a_single_item_should_still_be_buffered_even_if_it_is_above_the_buffer_size(self):
        input = [b'123456']
        result = list(data_collector(input, def_buf_size=3))
        self.assertEqual(result, [b'123', b'456'])

    def test_should_return_rest_of_data_on_last_iteration(self):
        input = [b'1234', b'56']
        result = list(data_collector(input, def_buf_size=3))
        self.assertEqual(result, [b'123', b'456'])

if __name__ == '__main__':
    unittest.main()



# Pansharpened Image Process using Rasterio
# Landsat Util
# License: CC0 1.0 Universal

from __future__ import print_function, division, absolute_import

import sys
import subprocess
from termcolor import colored


class VerbosityMixin(object):
    """
    Verbosity Mixin that generates beautiful stdout outputs.
    """

    verbose = False

    def output(self, value, normal=False, color=None, error=False,
               arrow=False, indent=None):
        """ Handles verbosity of this calls.
        if priority is set to 1, the value is printed

        if class instance verbose is True, the value is printed

        :param value:
            a string representing the message to be printed
        :type value:
            String
        :param normal:
            if set to true the message is always printed, otherwise it is only shown if verbosity is set
        :type normal:
            boolean
        :param color:
            The color of the message, choices: 'red', 'green', 'blue'
        :type normal:
            String
        :param error:
            if set to true the message appears in red
        :type error:
            Boolean
        :param arrow:
            if set to true an arrow appears before the message
        :type arrow:
            Boolean
        :param indent:
            indents the message based on the number provided
        :type indent:
            Boolean

        :returns:
            void
        """

        if error and value and (normal or self.verbose):
            return self._print(value, color='red', indent=indent)

        if self.verbose or normal:
            return self._print(value, color, arrow, indent)

        return

    def subprocess(self, argv):
        """
        Execute subprocess commands with proper ouput.
        This is no longer used in landsat-util

        :param argv:
            A list of subprocess arguments
        :type argv:
            List

        :returns:
            void
        """

        if self.verbose:
            proc = subprocess.Popen(argv, stderr=subprocess.PIPE)
        else:
            proc = subprocess.Popen(argv, stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)

        self.output(proc.stderr.read(), error=True)

        return

    def exit(self, message):
        """ outputs an exit message and exits

        :param message:
            The message to be outputed
        :type message:
            String

        :returns:
            void
        """

        self.output(message, normal=True, color="green")
        sys.exit()

    def _print(self, msg, color=None, arrow=False, indent=None):
        """ Print the msg with the color provided. """
        if color:
            msg = colored(msg, color)

        if arrow:
            msg = colored('===> ', 'blue') + msg

        if indent:
            msg = ('     ' * indent) + msg

        print(msg)

        return msg



#!/usr/bin/env python

# Landsat Util
# License: CC0 1.0 Universal

try:
    from setuptools import setup
    setup_kwargs = {'entry_points': {'console_scripts':['landsat=landsat.landsat:__main__']}}
except ImportError:
    from distutils.core import setup
    setup_kwargs = {'scripts': ['bin/landsat']}
    

from landsat import __version__


def readme():
    with open('README.rst') as f:
        return f.read()

with open('requirements.txt') as fid:
    INSTALL_REQUIRES = [l.strip() for l in fid.readlines() if l]

with open('requirements-dev.txt') as fid:
    TEST_REQUIRES = [l.strip() for l in fid.readlines() if l]

setup(
    name='landsat-util',
    version=__version__,
    description='A utility to search, download and process Landsat 8' +
    ' satellite imagery',
    long_description=readme(),
    author='Development Seed',
    author_email='info@developmentseed.org',
    url='https://github.com/developmentseed/landsat-util',
    packages=['landsat'],
    include_package_data=True,
    license='CCO',
    platforms='Posix; MacOS X; Windows',
    install_requires=INSTALL_REQUIRES,
    test_suite='nose.collector',
    tests_require=TEST_REQUIRES,
    **setup_kwargs
)



# Landsat Util
# License: CC0 1.0 Universal

"""Tests for search"""

import unittest

from jsonschema import validate

from landsat.search import Search
from tests import geojson_schema


class TestSearchHelper(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.s = Search()

    def test_search(self):
        # TEST A REGULAR SEARCH WITH KNOWN RESULT for paths and rows
        paths_rows = '003,003'
        start_date = '2014-01-01'
        end_date = '2014-06-01'

        result = self.s.search(paths_rows=paths_rows, start_date=start_date, end_date=end_date)
        self.assertEqual('2014-05-22', result['results'][0]['date'])

        # TEST A REGULAR SEARCH WITH KNOWN RESULT for lat and lon
        lat = 38.9107203
        lon = -77.0290116
        start_date = '2015-02-01'
        end_date = '2015-02-20'

        result = self.s.search(lat=lat, lon=lon, start_date=start_date, end_date=end_date)
        self.assertEqual('2015-02-06', result['results'][0]['date'])

    def test_search_zero_lon(self):
        # Make sure that zero coordinates are handled correctly
        paths_rows = '003,003'
        lon = 0.0
        lat = 52.0
        start_date = '2016-01-01'
        end_date = '2016-01-10'

        result = self.s.search(start_date=start_date, end_date=end_date,
                               lon=0.0, lat=52.0)
        self.assertEqual('2016-01-06', result['results'][0]['date'])

    def test_search_with_geojson(self):

        # TEST A REGULAR SEARCH WITH KNOWN RESULT for paths and rows
        paths_rows = '003,003'
        start_date = '2014-01-01'
        end_date = '2014-06-01'

        result = self.s.search(paths_rows=paths_rows, start_date=start_date, end_date=end_date, geojson=True)
        self.assertIsNone(validate(result, geojson_schema))
        self.assertEqual('2014-05-22', result['features'][0]['properties']['date'])

    def test_query_builder(self):
        # test wiht no input
        string = self.s.query_builder()
        self.assertEqual('', string)

        # just with rows and paths
        string = self.s.query_builder(paths_rows='003,004')
        self.assertEqual('(path:003+AND+row:004)', string)

        # multiple rows and paths
        string = self.s.query_builder(paths_rows='003,004,010,001')
        self.assertEqual('(path:003+AND+row:004)+OR+(path:010+AND+row:001)', string)

        # incomplete rows and paths
        self.assertRaises(ValueError, self.s.query_builder, paths_rows='003,004,010')

        # full example
        expected_string = ('acquisitionDate:[2014-01-01+TO+2014-11-12]+AND+cloudCoverFull:[10+TO+28]+AND+upperLeftCo'
                           'rnerLatitude:[23+TO+1000]+AND+lowerRightCornerLatitude:[-1000+TO+23]+AND+lowerLeftCorner'
                           'Longitude:[-1000+TO+21]+AND+upperRightCornerLongitude:[21+TO+1000]+AND+((path:003+AND+ro'
                           'w:004))')
        string = self.s.query_builder(paths_rows='003,004', lat=23, lon=21, start_date='2014-01-01',
                                      end_date='2014-11-12', cloud_min=10, cloud_max=28)
        self.assertEqual(expected_string, string)

    def test_lat_lon_builder(self):
        expected_string = ('upperLeftCornerLatitude:[12.3344+TO+1000]+AND+lowerRightCornerLatitude:[-1000+TO+12.3344]'
                           '+AND+lowerLeftCornerLongitude:[-1000+TO+11.0032]+AND+upperRightCornerLongitude:[11.0032+T'
                           'O+1000]')

        # Test with floats
        string = self.s.lat_lon_builder(12.3344, 11.0032)
        self.assertEqual(expected_string, string)

        # Test with strings
        string = self.s.lat_lon_builder('12.3344', '11.0032')
        self.assertEqual(expected_string, string)

    def test_cloud_cover_prct_range_builder(self):
        # no input
        string = self.s.cloud_cover_prct_range_builder()
        self.assertEqual('cloudCoverFull:[0+TO+100]', string)

        # just min
        string = self.s.cloud_cover_prct_range_builder(3)
        self.assertEqual('cloudCoverFull:[3+TO+100]', string)

        # just max
        string = self.s.cloud_cover_prct_range_builder(max=30)
        self.assertEqual('cloudCoverFull:[0+TO+30]', string)

        # both inputs
        string = self.s.cloud_cover_prct_range_builder(7, 10)
        self.assertEqual('cloudCoverFull:[7+TO+10]', string)

    def test_date_range_builder(self):
        string = self.s.date_range_builder('2014-01-01', '2015-01-01')
        self.assertEqual('acquisitionDate:[2014-01-01+TO+2015-01-01]', string)

    def test_row_path_builder(self):
        string = self.s.row_path_builder('003', '004')
        self.assertEqual('path:003+AND+row:004', string)


if __name__ == '__main__':
    unittest.main()



# Pansharpened Image Process using Rasterio
# Landsat Util
# License: CC0 1.0 Universal

from __future__ import print_function, division, absolute_import

import os
import tarfile
import glob
from copy import copy
import subprocess
from shutil import copyfile
from os.path import join, isdir

import numpy
import rasterio
from rasterio.coords import disjoint_bounds
from rasterio.warp import reproject, RESAMPLING, transform, transform_bounds

from skimage import transform as sktransform
from skimage.util import img_as_ubyte
from skimage.exposure import rescale_intensity
from polyline.codec import PolylineCodec

from .mixins import VerbosityMixin
from .utils import get_file, check_create_folder, exit, adjust_bounding_box
from .decorators import rasterio_decorator


class FileDoesNotExist(Exception):
    """ Exception to be used when the file does not exist. """
    pass


class BoundsDoNotOverlap(Exception):
    """ Exception for when bounds do not overlap with the image """
    pass


class BaseProcess(VerbosityMixin):
    """
    Image procssing class

    To initiate the following parameters must be passed:

    :param path:
        Path of the image.
    :type path:
        String
    :param bands:
        The band sequence for the final image. Must be a python list. (optional)
    :type bands:
        List
    :param dst_path:
        Path to the folder where the image should be stored. (optional)
    :type dst_path:
        String
    :param verbose:
        Whether the output should be verbose. Default is False.
    :type verbose:
        boolean
    :param force_unzip:
        Whether to force unzip the tar file. Default is False
    :type force_unzip:
        boolean

    """

    def __init__(self, path, bands=None, dst_path=None, verbose=False, force_unzip=False, bounds=None):

        self.projection = {'init': 'epsg:3857'}
        self.dst_crs = {'init': u'epsg:3857'}
        self.scene = get_file(path).split('.')[0]
        self.bands = bands if isinstance(bands, list) else [4, 3, 2]
        self.clipped = False

        # Landsat source path
        self.src_path = path.replace(get_file(path), '')

        # Build destination folder if doesn't exist
        self.dst_path = dst_path if dst_path else os.getcwd()
        self.dst_path = check_create_folder(join(self.dst_path, self.scene))
        self.verbose = verbose

        # Path to the unzipped folder
        self.scene_path = join(self.src_path, self.scene)

        # Unzip files
        if self._check_if_zipped(path):
            self._unzip(join(self.src_path, get_file(path)), join(self.src_path, self.scene), self.scene, force_unzip)

        if (bounds):
            self.bounds = bounds
            self.scene_path = self.clip()
            self.clipped = True

        self.bands_path = []
        for band in self.bands:
            self.bands_path.append(join(self.scene_path, self._get_full_filename(band)))

    def _get_boundaries(self, src, shape):

        self.output("Getting boundaries", normal=True, arrow=True)
        output = {'ul': {'x': [0, 0], 'y': [0, 0]},  # ul: upper left
                  'ur': {'x': [0, 0], 'y': [0, 0]},  # ur: upper right
                  'll': {'x': [0, 0], 'y': [0, 0]},  # ll: lower left
                  'lr': {'x': [0, 0], 'y': [0, 0]}}  # lr: lower right

        output['ul']['x'][0] = src['affine'][2]
        output['ul']['y'][0] = src['affine'][5]
        output['ur']['x'][0] = output['ul']['x'][0] + self.pixel * src['shape'][1]
        output['ur']['y'][0] = output['ul']['y'][0]
        output['ll']['x'][0] = output['ul']['x'][0]
        output['ll']['y'][0] = output['ul']['y'][0] - self.pixel * src['shape'][0]
        output['lr']['x'][0] = output['ul']['x'][0] + self.pixel * src['shape'][1]
        output['lr']['y'][0] = output['ul']['y'][0] - self.pixel * src['shape'][0]

        output['ul']['x'][1], output['ul']['y'][1] = transform(src['crs'], self.projection,
                                                               [output['ul']['x'][0]],
                                                               [output['ul']['y'][0]])

        output['ur']['x'][1], output['ur']['y'][1] = transform(src['crs'], self.projection,
                                                               [output['ur']['x'][0]],
                                                               [output['ur']['y'][0]])

        output['ll']['x'][1], output['ll']['y'][1] = transform(src['crs'], self.projection,
                                                               [output['ll']['x'][0]],
                                                               [output['ll']['y'][0]])

        output['lr']['x'][1], output['lr']['y'][1] = transform(src['crs'], self.projection,
                                                               [output['lr']['x'][0]],
                                                               [output['lr']['y'][0]])

        dst_corner_ys = [output[k]['y'][1][0] for k in output.keys()]
        dst_corner_xs = [output[k]['x'][1][0] for k in output.keys()]
        y_pixel = abs(max(dst_corner_ys) - min(dst_corner_ys)) / shape[0]
        x_pixel = abs(max(dst_corner_xs) - min(dst_corner_xs)) / shape[1]

        return (min(dst_corner_xs), x_pixel, 0.0, max(dst_corner_ys), 0.0, -y_pixel)

    def _read_bands(self):
        """ Reads a band with rasterio """
        bands = []

        try:
            for i, band in enumerate(self.bands):
                bands.append(rasterio.open(self.bands_path[i]).read_band(1))
        except IOError as e:
            exit(e.message, 1)

        return bands

    def _warp(self, proj_data, bands, new_bands):
        self.output("Projecting", normal=True, arrow=True)
        for i, band in enumerate(bands):
            self.output("band %s" % self.bands[i], normal=True, color='green', indent=1)
            reproject(band, new_bands[i], src_transform=proj_data['transform'], src_crs=proj_data['crs'],
                      dst_transform=proj_data['dst_transform'], dst_crs=self.dst_crs, resampling=RESAMPLING.nearest,
                      num_threads=2)

    def _unzip(self, src, dst, scene, force_unzip=False):
        """ Unzip tar files """
        self.output("Unzipping %s - It might take some time" % scene, normal=True, arrow=True)

        try:
            # check if file is already unzipped, skip
            if isdir(dst) and not force_unzip:
                self.output('%s is already unzipped.' % scene, normal=True, color='green', indent=1)
                return
            else:
                tar = tarfile.open(src, 'r')
                tar.extractall(path=dst)
                tar.close()
        except tarfile.ReadError:
            check_create_folder(dst)
            subprocess.check_call(['tar', '-xf', src, '-C', dst])

    def _get_full_filename(self, band):

        base_file = '%s_B%s.*' % (self.scene, band)

        try:
            return glob.glob(join(self.scene_path, base_file))[0].split('/')[-1]
        except IndexError:
            raise FileDoesNotExist('%s does not exist' % '%s_B%s.*' % (self.scene, band))

    def _check_if_zipped(self, path):
        """ Checks if the filename shows a tar/zip file """

        filename = get_file(path).split('.')

        if filename[-1] in ['bz', 'bz2', 'gz']:
            return True

        return False

    def _read_metadata(self):
        output = {}
        try:
            with open(self.scene_path + '/' + self.scene + '_MTL.txt', 'rU') as mtl:
                lines = mtl.readlines()
                for line in lines:
                    if 'REFLECTANCE_ADD_BAND_3' in line:
                        output['REFLECTANCE_ADD_BAND_3'] = float(line.replace('REFLECTANCE_ADD_BAND_3 = ', ''))

                    if 'REFLECTANCE_MULT_BAND_3' in line:
                        output['REFLECTANCE_MULT_BAND_3'] = float(line.replace('REFLECTANCE_MULT_BAND_3 = ', ''))

                    if 'REFLECTANCE_ADD_BAND_4' in line:
                        output['REFLECTANCE_ADD_BAND_4'] = float(line.replace('REFLECTANCE_ADD_BAND_4 = ', ''))

                    if 'REFLECTANCE_MULT_BAND_4' in line:
                        output['REFLECTANCE_MULT_BAND_4'] = float(line.replace('REFLECTANCE_MULT_BAND_4 = ', ''))

                    if 'CLOUD_COVER' in line:
                        output['CLOUD_COVER'] = float(line.replace('CLOUD_COVER = ', ''))

                    return output
        except IOError:
            return output

    def _get_image_data(self):
        src = rasterio.open(self.bands_path[-1])

        # Get pixel size from source
        self.pixel = src.affine[0]

        # Only collect src data that is needed and delete the rest
        image_data = {
            'transform': src.transform,
            'crs': src.crs,
            'affine': src.affine,
            'shape': src.shape,
            'dst_transform': None
        }

        image_data['dst_transform'] = self._get_boundaries(image_data, image_data['shape'])

        return image_data

    def _generate_new_bands(self, shape):
        new_bands = []
        for i in range(0, 3):
            new_bands.append(numpy.empty(shape, dtype=numpy.uint16))

        return new_bands

    @rasterio_decorator
    def _write_to_file(self, new_bands, **kwargs):

        # Read coverage from QBA
        coverage = self._calculate_cloud_ice_perc()

        self.output("Final Steps", normal=True, arrow=True)

        suffix = 'bands_%s' % "".join(map(str, self.bands))

        output_file = join(self.dst_path, self._filename(suffix=suffix))

        output = rasterio.open(output_file, 'w', **kwargs)

        for i, band in enumerate(new_bands):
            # Color Correction
            band = self._color_correction(band, self.bands[i], 0, coverage)

            output.write_band(i + 1, img_as_ubyte(band))

            new_bands[i] = None
        self.output("Writing to file", normal=True, color='green', indent=1)

        return output_file

    def _color_correction(self, band, band_id, low, coverage):
        if self.bands == [4, 5]:
            return band
        else:
            self.output("Color correcting band %s" % band_id, normal=True, color='green', indent=1)
            p_low, cloud_cut_low = self._percent_cut(band, low, 100 - (coverage * 3 / 4))
            temp = numpy.zeros(numpy.shape(band), dtype=numpy.uint16)
            cloud_divide = 65000 - coverage * 100
            mask = numpy.logical_and(band < cloud_cut_low, band > 0)
            temp[mask] = rescale_intensity(band[mask], in_range=(p_low, cloud_cut_low), out_range=(256, cloud_divide))
            temp[band >= cloud_cut_low] = rescale_intensity(band[band >= cloud_cut_low],
                                                            out_range=(cloud_divide, 65535))
            return temp

    def _percent_cut(self, color, low, high):
        return numpy.percentile(color[numpy.logical_and(color > 0, color < 65535)], (low, high))

    def _calculate_cloud_ice_perc(self):
        """ Return the percentage of pixels that are either cloud or snow with
        high confidence (> 67%).
        """
        self.output('Calculating cloud and snow coverage from QA band', normal=True, arrow=True)
        a = rasterio.open(join(self.scene_path, self._get_full_filename('QA'))).read_band(1)

        cloud_high_conf = int('1100000000000000', 2)
        snow_high_conf = int('0000110000000000', 2)
        fill_pixels = int('0000000000000001', 2)
        cloud_mask = numpy.bitwise_and(a, cloud_high_conf) == cloud_high_conf
        snow_mask = numpy.bitwise_and(a, snow_high_conf) == snow_high_conf
        fill_mask = numpy.bitwise_and(a, fill_pixels) == fill_pixels

        perc = numpy.true_divide(numpy.sum(cloud_mask | snow_mask),
                                 a.size - numpy.sum(fill_mask)) * 100.0
        self.output('cloud/snow coverage: %s' % round(perc, 2), indent=1, normal=True, color='green')
        return perc

    def _filename(self, name=None, suffix=None, prefix=None):
        """ File name generator for processed images """

        filename = ''

        if prefix:
            filename += str(prefix) + '_'

        if name:
            filename += str(name)
        else:
            filename += str(self.scene)

        if suffix:
            filename += '_' + str(suffix)

        if self.clipped:
            bounds = [tuple(self.bounds[0:2]), tuple(self.bounds[2:4])]
            polyline = PolylineCodec().encode(bounds)
            filename += '_clipped_' + polyline

        filename += '.TIF'

        return filename

    @rasterio_decorator
    def clip(self):
        """ Clip images based on bounds provided
        Implementation is borrowed from
        https://github.com/brendan-ward/rasterio/blob/e3687ce0ccf8ad92844c16d913a6482d5142cf48/rasterio/rio/convert.py
        """

        self.output("Clipping", normal=True)

        # create new folder for clipped images
        path = check_create_folder(join(self.scene_path, 'clipped'))

        try:
            temp_bands = copy(self.bands)
            temp_bands.append('QA')
            for i, band in enumerate(temp_bands):
                band_name = self._get_full_filename(band)
                band_path = join(self.scene_path, band_name)

                self.output("Band %s" % band, normal=True, color='green', indent=1)
                with rasterio.open(band_path) as src:
                    bounds = transform_bounds(
                        {
                            'proj': 'longlat',
                            'ellps': 'WGS84',
                            'datum': 'WGS84',
                            'no_defs': True
                        },
                        src.crs,
                        *self.bounds
                    )

                    if disjoint_bounds(bounds, src.bounds):
                        bounds = adjust_bounding_box(src.bounds, bounds)

                    window = src.window(*bounds)

                    out_kwargs = src.meta.copy()
                    out_kwargs.update({
                        'driver': 'GTiff',
                        'height': window[0][1] - window[0][0],
                        'width': window[1][1] - window[1][0],
                        'transform': src.window_transform(window)
                    })

                    with rasterio.open(join(path, band_name), 'w', **out_kwargs) as out:
                        out.write(src.read(window=window))

            # Copy MTL to the clipped folder
            copyfile(join(self.scene_path, self.scene + '_MTL.txt'), join(path, self.scene + '_MTL.txt'))

            return path

        except IOError as e:
            exit(e.message, 1)


class Simple(BaseProcess):

    @rasterio_decorator
    def run(self):
        """ Executes the image processing.

        :returns:
            (String) the path to the processed image
        """

        self.output('Image processing started for bands %s' % '-'.join(map(str, self.bands)), normal=True, arrow=True)

        bands = self._read_bands()
        image_data = self._get_image_data()

        new_bands = self._generate_new_bands(image_data['shape'])

        self._warp(image_data, bands, new_bands)

        # Bands are no longer needed
        del bands

        rasterio_options = {
            'driver': 'GTiff',
            'width': image_data['shape'][1],
            'height': image_data['shape'][0],
            'count': 3,
            'dtype': numpy.uint8,
            'nodata': 0,
            'transform': image_data['dst_transform'],
            'photometric': 'RGB',
            'crs': self.dst_crs
        }

        return self._write_to_file(new_bands, **rasterio_options)


class PanSharpen(BaseProcess):

    def __init__(self, path, bands=None, **kwargs):
        if bands:
            bands.append(8)
        else:
            bands = [4, 3, 2, 8]

        self.band8 = bands.index(8)
        super(PanSharpen, self).__init__(path, bands, **kwargs)

    @rasterio_decorator
    def run(self):
        """ Executes the pansharpen image processing.
        :returns:
            (String) the path to the processed image
        """

        self.output('PanSharpened Image processing started for bands %s' % '-'.join(map(str, self.bands)),
                    normal=True, arrow=True)

        bands = self._read_bands()
        image_data = self._get_image_data()

        new_bands = self._generate_new_bands(image_data['shape'])

        bands[:3] = self._rescale(bands[:3])
        new_bands.append(numpy.empty(image_data['shape'], dtype=numpy.uint16))

        self._warp(image_data, bands, new_bands)

        # Bands are no longer needed
        del bands

        # Calculate pan band
        pan = self._pansize(new_bands)
        del self.bands[self.band8]
        del new_bands[self.band8]

        rasterio_options = {
            'driver': 'GTiff',
            'width': image_data['shape'][1],
            'height': image_data['shape'][0],
            'count': 3,
            'dtype': numpy.uint8,
            'nodata': 0,
            'transform': image_data['dst_transform'],
            'photometric': 'RGB',
            'crs': self.dst_crs
        }

        return self._write_to_file(new_bands, pan, **rasterio_options)

    @rasterio_decorator
    def _write_to_file(self, new_bands, pan, **kwargs):

        # Read coverage from QBA
        coverage = self._calculate_cloud_ice_perc()

        self.output("Final Steps", normal=True, arrow=True)

        suffix = 'bands_%s_pan' % "".join(map(str, self.bands))

        output_file = join(self.dst_path, self._filename(suffix=suffix))

        output = rasterio.open(output_file, 'w', **kwargs)

        for i, band in enumerate(new_bands):
            # Color Correction
            band = numpy.multiply(band, pan)
            band = self._color_correction(band, self.bands[i], 0, coverage)

            output.write_band(i + 1, img_as_ubyte(band))

            new_bands[i] = None

        self.output("Writing to file", normal=True, color='green', indent=1)

        return output_file

    def _pansize(self, bands):

        self.output('Calculating Pan Ratio', normal=True, arrow=True)

        m = numpy.add(bands[0], bands[1])
        m = numpy.add(m, bands[2])
        pan = numpy.multiply(numpy.nan_to_num(numpy.true_divide(1, m)), bands[self.band8])

        return pan

    def _rescale(self, bands):
        """ Rescale bands """
        self.output("Rescaling", normal=True, arrow=True)

        for key, band in enumerate(bands):
            self.output("band %s" % self.bands[key], normal=True, color='green', indent=1)
            bands[key] = sktransform.rescale(band, 2)
            bands[key] = (bands[key] * 65535).astype('uint16')

        return bands

if __name__ == '__main__':

    p = PanSharpen('/Users/ajdevseed/Desktop/LC81950282014159LGN00')

    p.run()
